"""
Unit tests for the docker module
"""

import logging

import pytest
import salt.config
import salt.loader
import salt.modules.dockermod as docker_mod
import salt.utils.platform
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock, Mock, call, patch

log = logging.getLogger(__name__)

pytest.importorskip(
    "docker", reason="The python 'docker' package must be installed to run these tests"
)


@pytest.fixture
def configure_loader_modules():
    utils = salt.loader.utils(
        salt.config.DEFAULT_MINION_OPTS.copy(),
        whitelist=[
            "args",
            "docker",
            "json",
            "state",
            "thin",
            "systemd",
            "path",
            "platform",
        ],
    )
    return {
        docker_mod: {
            "__context__": {"docker.docker_version": ""},
            "__utils__": utils,
        }
    }


def test_failed_login():
    """
    Check that when docker.login failed a retcode other then 0
    is part of the return.
    """
    client = Mock()
    get_client_mock = MagicMock(return_value=client)
    ref_out = {"stdout": "", "stderr": "login failed", "retcode": 1}
    with patch.dict(
        docker_mod.__pillar__,
        {
            "docker-registries": {
                "portus.example.com:5000": {
                    "username": "admin",
                    "password": "linux12345",
                    "email": "tux@example.com",
                }
            }
        },
    ):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            dunder_salt = {
                "cmd.run_all": MagicMock(return_value=ref_out),
                "config.get": MagicMock(return_value={}),
                "config.option": MagicMock(return_value={}),
            }
            with patch.dict(docker_mod.__salt__, dunder_salt):
                ret = docker_mod.login("portus.example.com:5000")
                assert "retcode" in ret
                assert ret["retcode"] != 0


def test_logout_calls_docker_cli_logout_single():
    client = Mock()
    get_client_mock = MagicMock(return_value=client)
    ref_out = {"stdout": "", "stderr": "", "retcode": 0}
    registry_auth_data = {
        "portus.example.com:5000": {
            "username": "admin",
            "password": "linux12345",
            "email": "tux@example.com",
        }
    }
    docker_mock = MagicMock(return_value=ref_out)
    with patch.object(docker_mod, "_get_client", get_client_mock):
        dunder_salt = {
            "config.get": MagicMock(return_value=registry_auth_data),
            "cmd.run_all": docker_mock,
            "config.option": MagicMock(return_value={}),
        }
        with patch.dict(docker_mod.__salt__, dunder_salt):
            ret = docker_mod.logout("portus.example.com:5000")
            assert "retcode" in ret
            assert ret["retcode"] == 0
            docker_mock.assert_called_with(
                ["docker", "logout", "portus.example.com:5000"],
                python_shell=False,
                output_loglevel="quiet",
            )


def test_logout_calls_docker_cli_logout_all():
    client = Mock()
    get_client_mock = MagicMock(return_value=client)
    ref_out = {"stdout": "", "stderr": "", "retcode": 0}
    registry_auth_data = {
        "portus.example.com:5000": {
            "username": "admin",
            "password": "linux12345",
            "email": "tux@example.com",
        },
        "portus2.example.com:5000": {
            "username": "admin",
            "password": "linux12345",
            "email": "tux@example.com",
        },
    }

    docker_mock = MagicMock(return_value=ref_out)
    with patch.object(docker_mod, "_get_client", get_client_mock):
        dunder_salt = {
            "config.get": MagicMock(return_value=registry_auth_data),
            "cmd.run_all": docker_mock,
            "config.option": MagicMock(return_value={}),
        }
        with patch.dict(docker_mod.__salt__, dunder_salt):
            ret = docker_mod.logout()
            assert "retcode" in ret
            assert ret["retcode"] == 0
            assert docker_mock.call_count == 2


def test_ps_with_host_true():
    """
    Check that docker.ps called with host is ``True``,
    include resutlt of ``network.interfaces`` command in returned result.
    """
    client = Mock()
    client.containers = MagicMock(return_value=[])
    get_client_mock = MagicMock(return_value=client)
    network_interfaces = Mock(return_value={"mocked": None})

    with patch.dict(docker_mod.__salt__, {"network.interfaces": network_interfaces}):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            ret = docker_mod.ps_(host=True)
            assert ret == {"host": {"interfaces": {"mocked": None}}}


def test_ps_with_filters():
    """
    Check that docker.ps accept filters parameter.
    """
    client = Mock()
    client.containers = MagicMock(return_value=[])
    get_client_mock = MagicMock(return_value=client)

    with patch.object(docker_mod, "_get_client", get_client_mock):
        docker_mod.ps_(filters={"label": "KEY"})
        client.containers.assert_called_once_with(all=True, filters={"label": "KEY"})


@pytest.mark.slow_test
@pytest.mark.parametrize(
    "command_name, args",
    (
        ("create", ()),
        ("rm_", ()),
        ("kill", ()),
        ("pause", ()),
        ("signal_", ("KILL",)),
        ("start_", ()),
        ("stop", ()),
        ("unpause", ()),
        ("_run", ("command",)),
        ("_script", ("command",)),
    ),
)
def test_check_mine_cache_is_refreshed_on_container_change_event(command_name, args):
    """
    Every command that might modify docker containers state.
    Should trig an update on ``mine.send``
    """
    if salt.utils.platform.is_windows() and command_name == "_script":
        pytest.skip(
            "Skipping test since `dockermod._script` will create a temporary file in /tmp "
            "which does not exist on windows"
        )
    with patch.object(docker_mod, "_get_exec_driver"):
        client_args_mock = MagicMock(
            return_value={
                "create_container": [
                    "image",
                    "command",
                    "hostname",
                    "user",
                    "detach",
                    "stdin_open",
                    "tty",
                    "ports",
                    "environment",
                    "volumes",
                    "network_disabled",
                    "name",
                    "entrypoint",
                    "working_dir",
                    "domainname",
                    "cpuset",
                    "host_config",
                    "mac_address",
                    "labels",
                    "volume_driver",
                    "stop_signal",
                    "networking_config",
                    "healthcheck",
                    "stop_timeout",
                ],
                "host_config": [
                    "binds",
                    "port_bindings",
                    "lxc_conf",
                    "publish_all_ports",
                    "links",
                    "privileged",
                    "dns",
                    "dns_search",
                    "volumes_from",
                    "network_mode",
                    "restart_policy",
                    "cap_add",
                    "cap_drop",
                    "devices",
                    "extra_hosts",
                    "read_only",
                    "pid_mode",
                    "ipc_mode",
                    "security_opt",
                    "ulimits",
                    "log_config",
                    "mem_limit",
                    "memswap_limit",
                    "mem_reservation",
                    "kernel_memory",
                    "mem_swappiness",
                    "cgroup_parent",
                    "group_add",
                    "cpu_quota",
                    "cpu_period",
                    "blkio_weight",
                    "blkio_weight_device",
                    "device_read_bps",
                    "device_write_bps",
                    "device_read_iops",
                    "device_write_iops",
                    "oom_kill_disable",
                    "shm_size",
                    "sysctls",
                    "tmpfs",
                    "oom_score_adj",
                    "dns_opt",
                    "cpu_shares",
                    "cpuset_cpus",
                    "userns_mode",
                    "pids_limit",
                    "isolation",
                    "auto_remove",
                    "storage_opt",
                ],
                "networking_config": [
                    "aliases",
                    "links",
                    "ipv4_address",
                    "ipv6_address",
                    "link_local_ips",
                ],
            }
        )

        mine_send = Mock()
        command = getattr(docker_mod, command_name)
        client = MagicMock()
        client.api_version = "1.12"
        with patch.dict(
            docker_mod.__salt__,
            {
                "mine.send": mine_send,
                "container_resource.run": MagicMock(),
                "config.get": MagicMock(return_value=True),
                "config.option": MagicMock(return_value=True),
                "cp.cache_file": MagicMock(return_value=False),
            },
        ):
            with patch.dict(
                docker_mod.__utils__,
                {"docker.get_client_args": client_args_mock},
            ):
                with patch.object(docker_mod, "_get_client", client):
                    command("container", *args)
        try:
            mine_send.assert_called_with("docker.ps", verbose=True, all=True, host=True)
        except AssertionError as exc:
            raise Exception(
                "command '{}' did not call docker.ps with expected "
                "arguments: {}".format(command_name, exc)
            )


def test_update_mine():
    """
    Test the docker.update_mine config option
    """

    def config_get_disabled(val, default):
        return {
            "base_url": docker_mod.NOTSET,
            "version": docker_mod.NOTSET,
            "docker.url": docker_mod.NOTSET,
            "docker.version": docker_mod.NOTSET,
            "docker.machine": docker_mod.NOTSET,
            "docker.update_mine": False,
        }[val]

    def config_get_enabled(val, default):
        return {
            "base_url": docker_mod.NOTSET,
            "version": docker_mod.NOTSET,
            "docker.url": docker_mod.NOTSET,
            "docker.version": docker_mod.NOTSET,
            "docker.machine": docker_mod.NOTSET,
            "docker.update_mine": True,
        }[val]

    mine_mock = Mock()
    dunder_salt = {
        "config.get": MagicMock(side_effect=config_get_disabled),
        "config.option": MagicMock(return_value=False),
        "mine.send": mine_mock,
    }
    with patch.dict(docker_mod.__salt__, dunder_salt), patch.dict(
        docker_mod.__context__, {"docker.client": Mock()}
    ), patch.object(docker_mod, "state", MagicMock(return_value="stopped")):
        docker_mod.stop("foo", timeout=1)
        mine_mock.assert_not_called()

    with patch.dict(docker_mod.__salt__, dunder_salt), patch.dict(
        docker_mod.__context__, {"docker.client": Mock()}
    ), patch.object(docker_mod, "state", MagicMock(return_value="stopped")):
        dunder_salt["config.get"].side_effect = config_get_enabled
        dunder_salt["config.option"].return_value = True
        docker_mod.stop("foo", timeout=1)
        mine_mock.assert_called_once()


@pytest.mark.skipif(
    docker_mod.docker.version_info < (1, 5, 0),
    reason="docker module must be installed to run this test or is too old. >=1.5.0",
)
def test_list_networks():
    """
    test list networks.
    """
    __salt__ = {
        "config.get": Mock(),
        "mine.send": Mock(),
    }
    client = Mock()
    client.api_version = "1.21"
    client.networks = Mock(
        return_value=[{"Name": "foo", "Id": "01234", "Containers": {}}]
    )
    get_client_mock = MagicMock(return_value=client)
    with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod.networks(names=["foo"], ids=["01234"])
    client.networks.assert_called_once_with(names=["foo"], ids=["01234"])


@pytest.mark.skipif(
    docker_mod.docker.version_info < (1, 5, 0),
    reason="docker module must be installed to run this test or is too old. >=1.5.0",
)
def test_create_network():
    """
    test create network.
    """
    __salt__ = {
        "config.get": Mock(),
        "mine.send": Mock(),
    }
    client = Mock()
    client.api_version = "1.21"
    get_client_mock = MagicMock(return_value=client)

    with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod.create_network(
                "foo",
                driver="bridge",
                driver_opts={},
                gateway="192.168.0.1",
                ip_range="192.168.0.128/25",
                subnet="192.168.0.0/24",
            )
    client.create_network.assert_called_once_with(
        "foo",
        driver="bridge",
        options={},
        ipam={
            "Config": [
                {
                    "Gateway": "192.168.0.1",
                    "IPRange": "192.168.0.128/25",
                    "Subnet": "192.168.0.0/24",
                }
            ],
            "Driver": "default",
        },
        check_duplicate=True,
    )


@pytest.mark.skipif(
    docker_mod.docker.version_info < (1, 5, 0),
    reason="docker module must be installed to run this test or is too old. >=1.5.0",
)
def test_remove_network():
    """
    test remove network.
    """
    __salt__ = {
        "config.get": Mock(),
        "mine.send": Mock(),
    }
    client = Mock()
    client.api_version = "1.21"
    get_client_mock = MagicMock(return_value=client)

    with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod.remove_network("foo")
    client.remove_network.assert_called_once_with("foo")


@pytest.mark.skipif(
    docker_mod.docker.version_info < (1, 5, 0),
    reason="docker module must be installed to run this test or is too old. >=1.5.0",
)
def test_inspect_network():
    """
    test inspect network.
    """
    __salt__ = {
        "config.get": Mock(),
        "mine.send": Mock(),
    }
    client = Mock()
    client.api_version = "1.21"
    get_client_mock = MagicMock(return_value=client)

    with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod.inspect_network("foo")
    client.inspect_network.assert_called_once_with("foo")


@pytest.mark.skipif(
    docker_mod.docker.version_info < (1, 5, 0),
    reason="docker module must be installed to run this test or is too old. >=1.5.0",
)
def test_connect_container_to_network():
    """
    test connect_container_to_network
    """
    __salt__ = {
        "config.get": Mock(),
        "mine.send": Mock(),
    }
    client = Mock()
    client.api_version = "1.21"
    get_client_mock = MagicMock(return_value=client)

    context = {"docker.exec_driver": "docker-exec"}

    with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
        with patch.dict(docker_mod.__context__, context):
            with patch.object(docker_mod, "_get_client", get_client_mock):
                docker_mod.connect_container_to_network("container", "foo")
    client.connect_container_to_network.assert_called_once_with("container", "foo")


@pytest.mark.skipif(
    docker_mod.docker.version_info < (1, 5, 0),
    reason="docker module must be installed to run this test or is too old. >=1.5.0",
)
def test_disconnect_container_from_network():
    """
    test disconnect_container_from_network
    """
    __salt__ = {
        "config.get": Mock(),
        "mine.send": Mock(),
    }
    client = Mock()
    client.api_version = "1.21"
    get_client_mock = MagicMock(return_value=client)

    with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod.disconnect_container_from_network("container", "foo")
    client.disconnect_container_from_network.assert_called_once_with("container", "foo")


@pytest.mark.skipif(
    docker_mod.docker.version_info < (1, 5, 0),
    reason="docker module must be installed to run this test or is too old. >=1.5.0",
)
def test_list_volumes():
    """
    test list volumes.
    """
    __salt__ = {
        "config.get": Mock(),
        "mine.send": Mock(),
    }
    client = Mock()
    client.api_version = "1.21"
    get_client_mock = MagicMock(return_value=client)

    with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod.volumes(
                filters={"dangling": [True]},
            )
    client.volumes.assert_called_once_with(
        filters={"dangling": [True]},
    )


@pytest.mark.skipif(
    docker_mod.docker.version_info < (1, 5, 0),
    reason="docker module must be installed to run this test or is too old. >=1.5.0",
)
def test_create_volume():
    """
    test create volume.
    """
    __salt__ = {
        "config.get": Mock(),
        "mine.send": Mock(),
    }
    client = Mock()
    client.api_version = "1.21"
    get_client_mock = MagicMock(return_value=client)

    with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod.create_volume(
                "foo",
                driver="bridge",
                driver_opts={},
            )
    client.create_volume.assert_called_once_with(
        "foo",
        driver="bridge",
        driver_opts={},
    )


@pytest.mark.skipif(
    docker_mod.docker.version_info < (1, 5, 0),
    reason="docker module must be installed to run this test or is too old. >=1.5.0",
)
def test_remove_volume():
    """
    test remove volume.
    """
    __salt__ = {
        "config.get": Mock(),
        "mine.send": Mock(),
    }
    client = Mock()
    client.api_version = "1.21"
    get_client_mock = MagicMock(return_value=client)

    with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod.remove_volume("foo")
    client.remove_volume.assert_called_once_with("foo")


@pytest.mark.skipif(
    docker_mod.docker.version_info < (1, 5, 0),
    reason="docker module must be installed to run this test or is too old. >=1.5.0",
)
def test_inspect_volume():
    """
    test inspect volume.
    """
    __salt__ = {
        "config.get": Mock(),
        "mine.send": Mock(),
    }
    client = Mock()
    client.api_version = "1.21"
    get_client_mock = MagicMock(return_value=client)

    with patch.dict(docker_mod.__dict__, {"__salt__": __salt__}):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod.inspect_volume("foo")
    client.inspect_volume.assert_called_once_with("foo")


def test_wait_success():
    client = Mock()
    client.api_version = "1.21"
    client.wait = Mock(return_value=0)
    get_client_mock = MagicMock(return_value=client)

    docker_inspect_container = Mock(
        side_effect=[{"State": {"Running": True}}, {"State": {"Stopped": True}}]
    )
    with patch.object(docker_mod, "inspect_container", docker_inspect_container):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod._clear_context()
            result = docker_mod.wait("foo")
    assert result == {
        "result": True,
        "exit_status": 0,
        "state": {"new": "stopped", "old": "running"},
    }


def test_wait_fails_already_stopped():
    client = Mock()
    client.api_version = "1.21"
    client.wait = Mock(return_value=0)
    get_client_mock = MagicMock(return_value=client)

    docker_inspect_container = Mock(
        side_effect=[{"State": {"Stopped": True}}, {"State": {"Stopped": True}}]
    )
    with patch.object(docker_mod, "inspect_container", docker_inspect_container):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod._clear_context()
            result = docker_mod.wait("foo")
    assert result == {
        "result": False,
        "comment": "Container 'foo' already stopped",
        "exit_status": 0,
        "state": {"new": "stopped", "old": "stopped"},
    }


def test_wait_success_already_stopped():
    client = Mock()
    client.api_version = "1.21"
    client.wait = Mock(return_value=0)
    get_client_mock = MagicMock(return_value=client)

    docker_inspect_container = Mock(
        side_effect=[{"State": {"Stopped": True}}, {"State": {"Stopped": True}}]
    )
    with patch.object(docker_mod, "inspect_container", docker_inspect_container):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod._clear_context()
            result = docker_mod.wait("foo", ignore_already_stopped=True)
    assert result == {
        "result": True,
        "comment": "Container 'foo' already stopped",
        "exit_status": 0,
        "state": {"new": "stopped", "old": "stopped"},
    }


def test_wait_success_absent_container():
    client = Mock()
    client.api_version = "1.21"
    get_client_mock = MagicMock(return_value=client)

    docker_inspect_container = Mock(side_effect=CommandExecutionError)
    with patch.object(docker_mod, "inspect_container", docker_inspect_container):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod._clear_context()
            result = docker_mod.wait("foo", ignore_already_stopped=True)
    assert result == {"result": True, "comment": "Container 'foo' absent"}


def test_wait_fails_on_exit_status():
    client = Mock()
    client.api_version = "1.21"
    client.wait = Mock(return_value=1)
    get_client_mock = MagicMock(return_value=client)

    docker_inspect_container = Mock(
        side_effect=[{"State": {"Running": True}}, {"State": {"Stopped": True}}]
    )
    with patch.object(docker_mod, "inspect_container", docker_inspect_container):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod._clear_context()
            result = docker_mod.wait("foo", fail_on_exit_status=True)
    assert result == {
        "result": False,
        "exit_status": 1,
        "state": {"new": "stopped", "old": "running"},
    }


def test_wait_fails_on_exit_status_and_already_stopped():
    client = Mock()
    client.api_version = "1.21"
    client.wait = Mock(return_value=1)
    get_client_mock = MagicMock(return_value=client)

    docker_inspect_container = Mock(
        side_effect=[{"State": {"Stopped": True}}, {"State": {"Stopped": True}}]
    )
    with patch.object(docker_mod, "inspect_container", docker_inspect_container):
        with patch.object(docker_mod, "_get_client", get_client_mock):
            docker_mod._clear_context()
            result = docker_mod.wait(
                "foo", ignore_already_stopped=True, fail_on_exit_status=True
            )
    assert result == {
        "result": False,
        "comment": "Container 'foo' already stopped",
        "exit_status": 1,
        "state": {"new": "stopped", "old": "stopped"},
    }


def test_sls_build():
    """
    test build sls image.
    """
    docker_start_mock = MagicMock(return_value={})
    docker_create_mock = MagicMock(return_value={"Id": "ID", "Name": "NAME"})
    docker_stop_mock = MagicMock(
        return_value={"state": {"old": "running", "new": "stopped"}, "result": True}
    )
    docker_rm_mock = MagicMock(return_value={})
    docker_commit_mock = MagicMock(
        return_value={"Id": "ID2", "Image": "foo", "Time_Elapsed": 42}
    )

    docker_sls_mock = MagicMock(
        return_value={
            "file_|-/etc/test.sh_|-/etc/test.sh_|-managed": {
                "comment": "File /etc/test.sh is in the correct state",
                "name": "/etc/test.sh",
                "start_time": "07:04:26.834792",
                "result": True,
                "duration": 13.492,
                "__run_num__": 0,
                "changes": {},
            },
            "test_|-always-passes_|-foo_|-succeed_without_changes": {
                "comment": "Success!",
                "name": "foo",
                "start_time": "07:04:26.848915",
                "result": True,
                "duration": 0.363,
                "__run_num__": 1,
                "changes": {},
            },
        }
    )

    ret = None
    with patch.object(docker_mod, "start_", docker_start_mock), patch.object(
        docker_mod, "create", docker_create_mock
    ), patch.object(docker_mod, "stop", docker_stop_mock), patch.object(
        docker_mod, "commit", docker_commit_mock
    ), patch.object(
        docker_mod, "sls", docker_sls_mock
    ), patch.object(
        docker_mod, "rm_", docker_rm_mock
    ):
        ret = docker_mod.sls_build("foo", mods="foo")
    docker_create_mock.assert_called_once_with(
        cmd="sleep infinity", image="opensuse/python", interactive=True, tty=True
    )
    docker_start_mock.assert_called_once_with("ID")
    docker_sls_mock.assert_called_once_with("ID", "foo")
    docker_stop_mock.assert_called_once_with("ID")
    docker_rm_mock.assert_called_once_with("ID")
    docker_commit_mock.assert_called_once_with("ID", "foo", tag="latest")
    assert {"Id": "ID2", "Image": "foo", "Time_Elapsed": 42} == ret


def test_sls_build_dryrun():
    """
    test build sls image in dryrun mode.
    """
    docker_start_mock = MagicMock(return_value={})
    docker_create_mock = MagicMock(return_value={"Id": "ID", "Name": "NAME"})
    docker_stop_mock = MagicMock(
        return_value={"state": {"old": "running", "new": "stopped"}, "result": True}
    )
    docker_rm_mock = MagicMock(return_value={})

    docker_sls_mock = MagicMock(
        return_value={
            "file_|-/etc/test.sh_|-/etc/test.sh_|-managed": {
                "comment": "File /etc/test.sh is in the correct state",
                "name": "/etc/test.sh",
                "start_time": "07:04:26.834792",
                "result": True,
                "duration": 13.492,
                "__run_num__": 0,
                "changes": {},
            },
            "test_|-always-passes_|-foo_|-succeed_without_changes": {
                "comment": "Success!",
                "name": "foo",
                "start_time": "07:04:26.848915",
                "result": True,
                "duration": 0.363,
                "__run_num__": 1,
                "changes": {},
            },
        }
    )

    ret = None
    with patch.object(docker_mod, "start_", docker_start_mock), patch.object(
        docker_mod, "create", docker_create_mock
    ), patch.object(docker_mod, "stop", docker_stop_mock), patch.object(
        docker_mod, "rm_", docker_rm_mock
    ), patch.object(
        docker_mod, "sls", docker_sls_mock
    ):
        ret = docker_mod.sls_build("foo", mods="foo", dryrun=True)
    docker_create_mock.assert_called_once_with(
        cmd="sleep infinity", image="opensuse/python", interactive=True, tty=True
    )
    docker_start_mock.assert_called_once_with("ID")
    docker_sls_mock.assert_called_once_with("ID", "foo")
    docker_stop_mock.assert_called_once_with("ID")
    docker_rm_mock.assert_called_once_with("ID")
    assert {
        "file_|-/etc/test.sh_|-/etc/test.sh_|-managed": {
            "comment": "File /etc/test.sh is in the correct state",
            "name": "/etc/test.sh",
            "start_time": "07:04:26.834792",
            "result": True,
            "duration": 13.492,
            "__run_num__": 0,
            "changes": {},
        },
        "test_|-always-passes_|-foo_|-succeed_without_changes": {
            "comment": "Success!",
            "name": "foo",
            "start_time": "07:04:26.848915",
            "result": True,
            "duration": 0.363,
            "__run_num__": 1,
            "changes": {},
        },
    } == ret


@pytest.mark.slow_test
def test_call_success():
    """
    test module calling inside containers
    """
    ret = None
    docker_run_all_mock = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": '{"retcode": 0, "comment": "container cmd"}',
            "stderr": "err",
        }
    )
    docker_copy_to_mock = MagicMock(return_value={"retcode": 0})
    docker_config_mock = MagicMock(return_value="")
    client = Mock()
    client.put_archive = Mock()
    get_client_mock = MagicMock(return_value=client)

    context = {"docker.exec_driver": "docker-exec"}
    salt_dunder = {"config.option": docker_config_mock}

    with patch.object(docker_mod, "run_all", docker_run_all_mock), patch.object(
        docker_mod, "copy_to", docker_copy_to_mock
    ), patch.object(docker_mod, "_get_client", get_client_mock), patch.dict(
        docker_mod.__opts__, {"cachedir": "/tmp"}
    ), patch.dict(
        docker_mod.__salt__, salt_dunder
    ), patch.dict(
        docker_mod.__context__, context
    ):
        # call twice to verify tmp path later
        for i in range(2):
            ret = docker_mod.call("ID", "test.arg", 1, 2, arg1="val1")

    # Check that the directory is different each time
    # [ call(name, [args]), ...
    assert "mkdir" in docker_run_all_mock.mock_calls[0][1][1]
    assert "mkdir" in docker_run_all_mock.mock_calls[5][1][1]
    assert (
        docker_run_all_mock.mock_calls[0][1][1]
        != docker_run_all_mock.mock_calls[5][1][1]
    )

    assert "python3 --version" == docker_run_all_mock.mock_calls[1][1][1]

    assert "salt-call" in docker_run_all_mock.mock_calls[3][1][1]
    assert "salt-call" in docker_run_all_mock.mock_calls[8][1][1]
    assert (
        docker_run_all_mock.mock_calls[3][1][1]
        != docker_run_all_mock.mock_calls[8][1][1]
    )

    # check thin untar
    assert "tarfile" in docker_run_all_mock.mock_calls[2][1][1]
    assert "tarfile" in docker_run_all_mock.mock_calls[7][1][1]
    assert (
        docker_run_all_mock.mock_calls[2][1][1]
        != docker_run_all_mock.mock_calls[7][1][1]
    )

    # check directory cleanup
    assert "rm -rf" in docker_run_all_mock.mock_calls[4][1][1]
    assert "rm -rf" in docker_run_all_mock.mock_calls[9][1][1]
    assert (
        docker_run_all_mock.mock_calls[4][1][1]
        != docker_run_all_mock.mock_calls[9][1][1]
    )

    assert {"retcode": 0, "comment": "container cmd"} == ret


def test_images_with_empty_tags():
    """
    docker 1.12 reports also images without tags with `null`.
    """
    client = Mock()
    client.api_version = "1.24"
    client.images = Mock(
        return_value=[
            {"Id": "sha256:abcde", "RepoTags": None},
            {"Id": "sha256:abcdef"},
            {"Id": "sha256:abcdefg", "RepoTags": ["image:latest"]},
        ]
    )
    get_client_mock = MagicMock(return_value=client)

    with patch.object(docker_mod, "_get_client", get_client_mock):
        docker_mod._clear_context()
        result = docker_mod.images()
    assert result == {"sha256:abcdefg": {"RepoTags": ["image:latest"]}}


def test_compare_container_image_id_resolution():
    """
    Test comparing two containers when one's inspect output is an ID and
    not formatted in image:tag notation.
    """

    def _inspect_container_effect(id_):
        return {
            "container1": {
                "Config": {"Image": "realimage:latest"},
                "HostConfig": {},
            },
            "container2": {"Config": {"Image": "image_id"}, "HostConfig": {}},
        }[id_]

    def _inspect_image_effect(id_):
        return {
            "realimage:latest": {"Id": "image_id"},
            "image_id": {"Id": "image_id"},
        }[id_]

    inspect_container_mock = MagicMock(side_effect=_inspect_container_effect)
    inspect_image_mock = MagicMock(side_effect=_inspect_image_effect)

    with patch.object(docker_mod, "inspect_container", inspect_container_mock):
        with patch.object(docker_mod, "inspect_image", inspect_image_mock):
            ret = docker_mod.compare_containers("container1", "container2")
            assert ret == {}


def test_compare_container_ulimits_order():
    """
    Test comparing two containers when the order of the Ulimits HostConfig
    values are different, but the values are the same.
    """

    def _inspect_container_effect(id_):
        return {
            "container1": {
                "Config": {},
                "HostConfig": {
                    "Ulimits": [
                        {"Hard": -1, "Soft": -1, "Name": "core"},
                        {"Hard": 65536, "Soft": 65536, "Name": "nofile"},
                    ]
                },
            },
            "container2": {
                "Config": {},
                "HostConfig": {
                    "Ulimits": [
                        {"Hard": 65536, "Soft": 65536, "Name": "nofile"},
                        {"Hard": -1, "Soft": -1, "Name": "core"},
                    ]
                },
            },
        }[id_]

    inspect_container_mock = MagicMock(side_effect=_inspect_container_effect)

    with patch.object(docker_mod, "inspect_container", inspect_container_mock):
        # pylint: disable=not-callable
        ret = docker_mod.compare_container("container1", "container2")
        # pylint: enable=not-callable
        assert ret == {}


def test_compare_container_env_order():
    """
    Test comparing two containers when the order of the Env HostConfig
    values are different, but the values are the same.
    """

    def _inspect_container_effect(id_):
        return {
            "container1": {
                "Config": {},
                "HostConfig": {"Env": ["FOO=bar", "HELLO=world"]},
            },
            "container2": {
                "Config": {},
                "HostConfig": {"Env": ["HELLO=world", "FOO=bar"]},
            },
        }[id_]

    inspect_container_mock = MagicMock(side_effect=_inspect_container_effect)

    with patch.object(docker_mod, "inspect_container", inspect_container_mock):
        # pylint: disable=not-callable
        ret = docker_mod.compare_container("container1", "container2")
        # pylint: enable=not-callable
        assert ret == {}


def test_resolve_tag():
    """
    Test the resolve_tag function. It runs docker.insect_image on the image
    name passed and then looks for the RepoTags key in the result
    """
    id_ = "sha256:6ad733544a6317992a6fac4eb19fe1df577d4dec7529efec28a5bd0edad0fd30"
    tags = ["foo:latest", "foo:bar"]
    mock_tagged = MagicMock(return_value={"Id": id_, "RepoTags": tags})
    mock_untagged = MagicMock(return_value={"Id": id_, "RepoTags": []})
    mock_unexpected = MagicMock(return_value={"Id": id_})
    mock_not_found = MagicMock(side_effect=CommandExecutionError())

    with patch.object(docker_mod, "inspect_image", mock_tagged):
        assert docker_mod.resolve_tag("foo") == tags[0]
        assert docker_mod.resolve_tag("foo", all=True) == tags

    with patch.object(docker_mod, "inspect_image", mock_untagged):
        assert docker_mod.resolve_tag("foo") == id_
        assert docker_mod.resolve_tag("foo", all=True) == [id_]

    with patch.object(docker_mod, "inspect_image", mock_unexpected):
        assert docker_mod.resolve_tag("foo") == id_
        assert docker_mod.resolve_tag("foo", all=True) == [id_]

    with patch.object(docker_mod, "inspect_image", mock_not_found):
        assert docker_mod.resolve_tag("foo") is False
        assert docker_mod.resolve_tag("foo", all=True) is False


def test_prune():
    """
    Test the prune function
    """

    def _run(**kwargs):
        side_effect = kwargs.pop("side_effect", None)
        # No arguments passed, we should be pruning everything but volumes
        client = Mock()
        if side_effect is not None:
            client.side_effect = side_effect
        with patch.object(docker_mod, "_client_wrapper", client):
            docker_mod.prune(**kwargs)
        return client

    # Containers only, no filters
    client = _run(containers=True)
    client.assert_called_once_with("prune_containers", filters={})

    # Containers only, with filters
    client = _run(containers=True, until="24h", label="foo,bar=baz")
    client.assert_called_once_with(
        "prune_containers", filters={"until": ["24h"], "label": ["foo", "bar=baz"]}
    )

    # Images only, no filters
    client = _run(images=True)
    client.assert_called_once_with("prune_images", filters={})

    # Images only, with filters
    client = _run(images=True, dangling=True, until="24h", label="foo,bar=baz")
    client.assert_called_once_with(
        "prune_images",
        filters={"dangling": True, "until": ["24h"], "label": ["foo", "bar=baz"]},
    )

    # Networks only, no filters
    client = _run(networks=True)
    client.assert_called_once_with("prune_networks", filters={})

    # Networks only, with filters
    client = _run(networks=True, until="24h", label="foo,bar=baz")
    client.assert_called_once_with(
        "prune_networks", filters={"until": ["24h"], "label": ["foo", "bar=baz"]}
    )

    # Volumes only, no filters
    client = _run(system=False, volumes=True)
    client.assert_called_once_with("prune_volumes", filters={})

    # Volumes only, with filters
    client = _run(system=False, volumes=True, label="foo,bar=baz")
    client.assert_called_once_with(
        "prune_volumes", filters={"label": ["foo", "bar=baz"]}
    )

    # Containers and images, no filters
    client = _run(containers=True, images=True)
    assert client.call_args_list == [
        call("prune_containers", filters={}),
        call("prune_images", filters={}),
    ]

    # Containers and images, with filters
    client = _run(containers=True, images=True, until="24h", label="foo,bar=baz")
    assert client.call_args_list == [
        call(
            "prune_containers",
            filters={"until": ["24h"], "label": ["foo", "bar=baz"]},
        ),
        call(
            "prune_images",
            filters={"until": ["24h"], "label": ["foo", "bar=baz"]},
        ),
    ]

    # System, no volumes, no filters, assuming prune_build in docker-py
    client = _run(system=True)
    assert client.call_args_list == [
        call("prune_containers", filters={}),
        call("prune_images", filters={}),
        call("prune_networks", filters={}),
        call("prune_build", filters={}),
    ]

    # System, no volumes, no filters, assuming prune_build not in docker-py
    client = _run(
        system=True,
        side_effect=[None, None, None, SaltInvocationError(), None, None, None],
    )
    assert client.call_args_list == [
        call("prune_containers", filters={}),
        call("prune_images", filters={}),
        call("prune_networks", filters={}),
        call("prune_build", filters={}),
        call("_url", "/build/prune"),
        call("_post", None),
        call("_result", None, True),
    ]

    # System, no volumes, with filters, assuming prune_build in docker-py
    client = _run(system=True, label="foo,bar=baz")
    assert client.call_args_list == [
        call("prune_containers", filters={"label": ["foo", "bar=baz"]}),
        call("prune_images", filters={"label": ["foo", "bar=baz"]}),
        call("prune_networks", filters={"label": ["foo", "bar=baz"]}),
        call("prune_build", filters={"label": ["foo", "bar=baz"]}),
    ]

    # System, no volumes, with filters, assuming prune_build not in docker-py
    client = _run(
        system=True,
        label="foo,bar=baz",
        side_effect=[None, None, None, SaltInvocationError(), None, None, None],
    )
    assert client.call_args_list == [
        call("prune_containers", filters={"label": ["foo", "bar=baz"]}),
        call("prune_images", filters={"label": ["foo", "bar=baz"]}),
        call("prune_networks", filters={"label": ["foo", "bar=baz"]}),
        call("prune_build", filters={"label": ["foo", "bar=baz"]}),
        call("_url", "/build/prune"),
        call("_post", None),
        call("_result", None, True),
    ]

    # System and volumes, no filters, assuming prune_build in docker-py
    client = _run(system=True, volumes=True)
    assert client.call_args_list == [
        call("prune_containers", filters={}),
        call("prune_images", filters={}),
        call("prune_networks", filters={}),
        call("prune_build", filters={}),
        call("prune_volumes", filters={}),
    ]

    # System and volumes, no filters, assuming prune_build not in docker-py
    client = _run(
        system=True,
        volumes=True,
        side_effect=[
            None,
            None,
            None,
            SaltInvocationError(),
            None,
            None,
            None,
            None,
        ],
    )
    assert client.call_args_list == [
        call("prune_containers", filters={}),
        call("prune_images", filters={}),
        call("prune_networks", filters={}),
        call("prune_build", filters={}),
        call("_url", "/build/prune"),
        call("_post", None),
        call("_result", None, True),
        call("prune_volumes", filters={}),
    ]

    # System and volumes with filters, assuming prune_build in docker-py
    client = _run(system=True, volumes=True, label="foo,bar=baz")
    assert client.call_args_list == [
        call("prune_containers", filters={"label": ["foo", "bar=baz"]}),
        call("prune_images", filters={"label": ["foo", "bar=baz"]}),
        call("prune_networks", filters={"label": ["foo", "bar=baz"]}),
        call("prune_build", filters={"label": ["foo", "bar=baz"]}),
        call("prune_volumes", filters={"label": ["foo", "bar=baz"]}),
    ]

    # System and volumes, with filters, assuming prune_build not in docker-py
    client = _run(
        system=True,
        volumes=True,
        label="foo,bar=baz",
        side_effect=[
            None,
            None,
            None,
            SaltInvocationError(),
            None,
            None,
            None,
            None,
        ],
    )
    assert client.call_args_list == [
        call("prune_containers", filters={"label": ["foo", "bar=baz"]}),
        call("prune_images", filters={"label": ["foo", "bar=baz"]}),
        call("prune_networks", filters={"label": ["foo", "bar=baz"]}),
        call("prune_build", filters={"label": ["foo", "bar=baz"]}),
        call("_url", "/build/prune"),
        call("_post", None),
        call("_result", None, True),
        call("prune_volumes", filters={"label": ["foo", "bar=baz"]}),
    ]


def test_port():
    """
    Test docker.port function. Note that this test case does not test what
    happens when a specific container name is passed and that container
    does not exist. When that happens, the Docker API will just raise a 404
    error. Since we're using as side_effect to mock
    docker.inspect_container, it would be meaningless to code raising an
    exception into it and then test that we raised that exception.
    """
    ports = {
        "foo": {
            "5555/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32768"}],
            "6666/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32769"}],
        },
        "bar": {
            "4444/udp": [{"HostIp": "0.0.0.0", "HostPort": "32767"}],
            "5555/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32768"}],
            "6666/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32769"}],
        },
        "baz": {
            "5555/tcp": [{"HostIp": "0.0.0.0", "HostPort": "32768"}],
            "6666/udp": [{"HostIp": "0.0.0.0", "HostPort": "32769"}],
        },
    }
    list_mock = MagicMock(return_value=["bar", "baz", "foo"])
    inspect_mock = MagicMock(
        side_effect=lambda x: {"NetworkSettings": {"Ports": ports.get(x)}}
    )
    with patch.object(docker_mod, "list_containers", list_mock), patch.object(
        docker_mod, "inspect_container", inspect_mock
    ):

        # Test with specific container name
        ret = docker_mod.port("foo")
        assert ret == ports["foo"]

        # Test with specific container name and filtering on port
        ret = docker_mod.port("foo", private_port="5555/tcp")
        assert ret == {"5555/tcp": ports["foo"]["5555/tcp"]}

        # Test using pattern expression
        ret = docker_mod.port("ba*")
        assert ret == {"bar": ports["bar"], "baz": ports["baz"]}
        ret = docker_mod.port("ba?")
        assert ret == {"bar": ports["bar"], "baz": ports["baz"]}
        ret = docker_mod.port("ba[rz]")
        assert ret == {"bar": ports["bar"], "baz": ports["baz"]}

        # Test using pattern expression and port filtering
        ret = docker_mod.port("ba*", private_port="6666/tcp")
        assert ret == {"bar": {"6666/tcp": ports["bar"]["6666/tcp"]}, "baz": {}}
        ret = docker_mod.port("ba?", private_port="6666/tcp")
        assert ret == {"bar": {"6666/tcp": ports["bar"]["6666/tcp"]}, "baz": {}}
        ret = docker_mod.port("ba[rz]", private_port="6666/tcp")
        assert ret == {"bar": {"6666/tcp": ports["bar"]["6666/tcp"]}, "baz": {}}
        ret = docker_mod.port("*")
        assert ret == ports
        ret = docker_mod.port("*", private_port="5555/tcp")
        assert ret == {
            "foo": {"5555/tcp": ports["foo"]["5555/tcp"]},
            "bar": {"5555/tcp": ports["bar"]["5555/tcp"]},
            "baz": {"5555/tcp": ports["baz"]["5555/tcp"]},
        }
        ret = docker_mod.port("*", private_port=6666)
        assert ret == {
            "foo": {"6666/tcp": ports["foo"]["6666/tcp"]},
            "bar": {"6666/tcp": ports["bar"]["6666/tcp"]},
            "baz": {"6666/udp": ports["baz"]["6666/udp"]},
        }
        ret = docker_mod.port("*", private_port="6666/tcp")
        assert ret == {
            "foo": {"6666/tcp": ports["foo"]["6666/tcp"]},
            "bar": {"6666/tcp": ports["bar"]["6666/tcp"]},
            "baz": {},
        }



"""
Utils for the NAPALM modules and proxy.

.. seealso::

    - :mod:`NAPALM grains: select network devices based on their characteristics <salt.grains.napalm>`
    - :mod:`NET module: network basic features <salt.modules.napalm_network>`
    - :mod:`NTP operational and configuration management module <salt.modules.napalm_ntp>`
    - :mod:`BGP operational and configuration management module <salt.modules.napalm_bgp>`
    - :mod:`Routes details <salt.modules.napalm_route>`
    - :mod:`SNMP configuration module <salt.modules.napalm_snmp>`
    - :mod:`Users configuration management <salt.modules.napalm_users>`

.. versionadded:: 2017.7.0
"""


import copy
import importlib
import logging
import traceback
from functools import wraps

import salt.output
import salt.utils.args
import salt.utils.platform

try:
    # will try to import NAPALM
    # https://github.com/napalm-automation/napalm
    # pylint: disable=unused-import,no-name-in-module
    import napalm
    import napalm.base as napalm_base

    # pylint: enable=unused-import,no-name-in-module
    HAS_NAPALM = True
    try:
        NAPALM_MAJOR = int(napalm.__version__.split(".")[0])
    except AttributeError:
        NAPALM_MAJOR = 0
except ImportError:
    HAS_NAPALM = False

try:
    # try importing ConnectionClosedException
    # from napalm-base
    # this exception has been introduced only in version 0.24.0
    from napalm.base.exceptions import ConnectionClosedException

    HAS_CONN_CLOSED_EXC_CLASS = True
except ImportError:
    HAS_CONN_CLOSED_EXC_CLASS = False

log = logging.getLogger(__file__)


def is_proxy(opts):
    """
    Is this a NAPALM proxy?
    """
    return (
        salt.utils.platform.is_proxy()
        and opts.get("proxy", {}).get("proxytype") == "napalm"
    )


def is_always_alive(opts):
    """
    Is always alive required?
    """
    return opts.get("proxy", {}).get("always_alive", True)


def not_always_alive(opts):
    """
    Should this proxy be always alive?
    """
    return (is_proxy(opts) and not is_always_alive(opts)) or is_minion(opts)


def is_minion(opts):
    """
    Is this a NAPALM straight minion?
    """
    return not salt.utils.platform.is_proxy() and "napalm" in opts


def virtual(opts, virtualname, filename):
    """
    Returns the __virtual__.
    """
    if (HAS_NAPALM and NAPALM_MAJOR >= 2) and (is_proxy(opts) or is_minion(opts)):
        return virtualname
    else:
        return (
            False,
            '"{vname}"" {filename} cannot be loaded: '
            "NAPALM is not installed: ``pip install napalm``".format(
                vname=virtualname, filename="({filename})".format(filename=filename)
            ),
        )


def call(napalm_device, method, *args, **kwargs):
    """
    Calls arbitrary methods from the network driver instance.
    Please check the readthedocs_ page for the updated list of getters.

    .. _readthedocs: http://napalm.readthedocs.org/en/latest/support/index.html#getters-support-matrix

    method
        Specifies the name of the method to be called.

    *args
        Arguments.

    **kwargs
        More arguments.

    :return: A dictionary with three keys:

        * result (True/False): if the operation succeeded
        * out (object): returns the object as-is from the call
        * comment (string): provides more details in case the call failed
        * traceback (string): complete traceback in case of exception. \
        Please submit an issue including this traceback \
        on the `correct driver repo`_ and make sure to read the FAQ_

    .. _`correct driver repo`: https://github.com/napalm-automation/napalm/issues/new
    .. FAQ_: https://github.com/napalm-automation/napalm#faq

    Example:

    .. code-block:: python

        salt.utils.napalm.call(
            napalm_object,
            'cli',
            [
                'show version',
                'show chassis fan'
            ]
        )
    """
    result = False
    out = None
    opts = napalm_device.get("__opts__", {})
    retry = kwargs.pop("__retry", True)  # retry executing the task?
    force_reconnect = kwargs.get("force_reconnect", False)
    if force_reconnect:
        log.debug("Forced reconnection initiated")
        log.debug("The current opts (under the proxy key):")
        log.debug(opts["proxy"])
        opts["proxy"].update(**kwargs)
        log.debug("Updated to:")
        log.debug(opts["proxy"])
        napalm_device = get_device(opts)
    try:
        if not napalm_device.get("UP", False):
            raise Exception("not connected")
        # if connected will try to execute desired command
        kwargs_copy = {}
        kwargs_copy.update(kwargs)
        for karg, warg in kwargs_copy.items():
            # lets clear None arguments
            # to not be sent to NAPALM methods
            if warg is None:
                kwargs.pop(karg)
        out = getattr(napalm_device.get("DRIVER"), method)(*args, **kwargs)
        # calls the method with the specified parameters
        result = True
    except Exception as error:  # pylint: disable=broad-except
        # either not connected
        # either unable to execute the command
        hostname = napalm_device.get("HOSTNAME", "[unspecified hostname]")
        err_tb = (
            traceback.format_exc()
        )  # let's get the full traceback and display for debugging reasons.
        if isinstance(error, NotImplementedError):
            comment = (
                "{method} is not implemented for the NAPALM {driver} driver!".format(
                    method=method, driver=napalm_device.get("DRIVER_NAME")
                )
            )
        elif (
            retry
            and HAS_CONN_CLOSED_EXC_CLASS
            and isinstance(error, ConnectionClosedException)
        ):
            # Received disconection whilst executing the operation.
            # Instructed to retry (default behaviour)
            #   thus trying to re-establish the connection
            #   and re-execute the command
            #   if any of the operations (close, open, call) will rise again ConnectionClosedException
            #   it will fail loudly.
            kwargs["__retry"] = False  # do not attempt re-executing
            comment = "Disconnected from {device}. Trying to reconnect.".format(
                device=hostname
            )
            log.error(err_tb)
            log.error(comment)
            log.debug("Clearing the connection with %s", hostname)
            call(napalm_device, "close", __retry=False)  # safely close the connection
            # Make sure we don't leave any TCP connection open behind
            #   if we fail to close properly, we might not be able to access the
            log.debug("Re-opening the connection with %s", hostname)
            call(napalm_device, "open", __retry=False)
            log.debug("Connection re-opened with %s", hostname)
            log.debug("Re-executing %s", method)
            return call(napalm_device, method, *args, **kwargs)
            # If still not able to reconnect and execute the task,
            #   the proxy keepalive feature (if enabled) will attempt
            #   to reconnect.
            # If the device is using a SSH-based connection, the failure
            #   will also notify the paramiko transport and the `is_alive` flag
            #   is going to be set correctly.
            # More background: the network device may decide to disconnect,
            #   although the SSH session itself is alive and usable, the reason
            #   being the lack of activity on the CLI.
            #   Paramiko's keepalive doesn't help in this case, as the ServerAliveInterval
            #   are targeting the transport layer, whilst the device takes the decision
            #   when there isn't any activity on the CLI, thus at the application layer.
            #   Moreover, the disconnect is silent and paramiko's is_alive flag will
            #   continue to return True, although the connection is already unusable.
            #   For more info, see https://github.com/paramiko/paramiko/issues/813.
            #   But after a command fails, the `is_alive` flag becomes aware of these
            #   changes and will return False from there on. And this is how the
            #   Salt proxy keepalive helps: immediately after the first failure, it
            #   will know the state of the connection and will try reconnecting.
        else:
            comment = (
                'Cannot execute "{method}" on {device}{port} as {user}. Reason:'
                " {error}!".format(
                    device=napalm_device.get("HOSTNAME", "[unspecified hostname]"),
                    port=(
                        ":{port}".format(
                            port=napalm_device.get("OPTIONAL_ARGS", {}).get("port")
                        )
                        if napalm_device.get("OPTIONAL_ARGS", {}).get("port")
                        else ""
                    ),
                    user=napalm_device.get("USERNAME", ""),
                    method=method,
                    error=error,
                )
            )
        log.error(comment)
        log.error(err_tb)
        return {"out": {}, "result": False, "comment": comment, "traceback": err_tb}
    finally:
        if opts and not_always_alive(opts) and napalm_device.get("CLOSE", True):
            # either running in a not-always-alive proxy
            # either running in a regular minion
            # close the connection when the call is over
            # unless the CLOSE is explicitly set as False
            napalm_device["DRIVER"].close()
    return {"out": out, "result": result, "comment": ""}


def get_device_opts(opts, salt_obj=None):
    """
    Returns the options of the napalm device.
    :pram: opts
    :return: the network device opts
    """
    network_device = {}
    # by default, look in the proxy config details
    device_dict = opts.get("proxy", {}) if is_proxy(opts) else opts.get("napalm", {})
    if opts.get("proxy") or opts.get("napalm"):
        opts["multiprocessing"] = device_dict.get("multiprocessing", False)
        # Most NAPALM drivers are SSH-based, so multiprocessing should default to False.
        # But the user can be allows one to have a different value for the multiprocessing, which will
        #   override the opts.
    if not device_dict:
        # still not able to setup
        log.error(
            "Incorrect minion config. Please specify at least the napalm driver name!"
        )
    # either under the proxy hier, either under the napalm in the config file
    network_device["HOSTNAME"] = (
        device_dict.get("host")
        or device_dict.get("hostname")
        or device_dict.get("fqdn")
        or device_dict.get("ip")
    )
    network_device["USERNAME"] = device_dict.get("username") or device_dict.get("user")
    network_device["DRIVER_NAME"] = device_dict.get("driver") or device_dict.get("os")
    network_device["PASSWORD"] = (
        device_dict.get("passwd")
        or device_dict.get("password")
        or device_dict.get("pass")
        or ""
    )
    network_device["TIMEOUT"] = device_dict.get("timeout", 60)
    network_device["OPTIONAL_ARGS"] = device_dict.get("optional_args", {})
    network_device["ALWAYS_ALIVE"] = device_dict.get("always_alive", True)
    network_device["PROVIDER"] = device_dict.get("provider")
    network_device["UP"] = False
    # get driver object form NAPALM
    if "config_lock" not in network_device["OPTIONAL_ARGS"]:
        network_device["OPTIONAL_ARGS"]["config_lock"] = False
    if (
        network_device["ALWAYS_ALIVE"]
        and "keepalive" not in network_device["OPTIONAL_ARGS"]
    ):
        network_device["OPTIONAL_ARGS"]["keepalive"] = 5  # 5 seconds keepalive
    return network_device


def get_device(opts, salt_obj=None):
    """
    Initialise the connection with the network device through NAPALM.
    :param: opts
    :return: the network device object
    """
    log.debug("Setting up NAPALM connection")
    network_device = get_device_opts(opts, salt_obj=salt_obj)
    provider_lib = napalm_base
    if network_device.get("PROVIDER"):
        # In case the user requires a different provider library,
        #   other than napalm-base.
        # For example, if napalm-base does not satisfy the requirements
        #   and needs to be enahanced with more specific features,
        #   we may need to define a custom library on top of napalm-base
        #   with the constraint that it still needs to provide the
        #   `get_network_driver` function. However, even this can be
        #   extended later, if really needed.
        # Configuration example:
        #   provider: napalm_base_example
        try:
            provider_lib = importlib.import_module(network_device.get("PROVIDER"))
        except ImportError as ierr:
            log.error(
                "Unable to import %s", network_device.get("PROVIDER"), exc_info=True
            )
            log.error("Falling back to napalm-base")
    _driver_ = provider_lib.get_network_driver(network_device.get("DRIVER_NAME"))
    try:
        network_device["DRIVER"] = _driver_(
            network_device.get("HOSTNAME", ""),
            network_device.get("USERNAME", ""),
            network_device.get("PASSWORD", ""),
            timeout=network_device["TIMEOUT"],
            optional_args=network_device["OPTIONAL_ARGS"],
        )
        network_device.get("DRIVER").open()
        # no exception raised here, means connection established
        network_device["UP"] = True
    except napalm_base.exceptions.ConnectionException as error:
        base_err_msg = "Cannot connect to {hostname}{port} as {username}.".format(
            hostname=network_device.get("HOSTNAME", "[unspecified hostname]"),
            port=(
                ":{port}".format(
                    port=network_device.get("OPTIONAL_ARGS", {}).get("port")
                )
                if network_device.get("OPTIONAL_ARGS", {}).get("port")
                else ""
            ),
            username=network_device.get("USERNAME", ""),
        )
        log.error(base_err_msg)
        log.error("Please check error: %s", error)
        raise napalm_base.exceptions.ConnectionException(base_err_msg)
    return network_device


def proxy_napalm_wrap(func):
    """
    This decorator is used to make the execution module functions
    available outside a proxy minion, or when running inside a proxy
    minion. If we are running in a proxy, retrieve the connection details
    from the __proxy__ injected variable.  If we are not, then
    use the connection information from the opts.
    :param func:
    :return:
    """

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        wrapped_global_namespace = func.__globals__
        # get __opts__ and __proxy__ from func_globals
        proxy = wrapped_global_namespace.get("__proxy__")
        opts = copy.deepcopy(wrapped_global_namespace.get("__opts__"))
        # in any case, will inject the `napalm_device` global
        # the execution modules will make use of this variable from now on
        # previously they were accessing the device properties through the __proxy__ object
        always_alive = opts.get("proxy", {}).get("always_alive", True)
        # force_reconnect is a magic keyword arg that allows one to establish
        # a separate connection to the network device running under an always
        # alive Proxy Minion, using new credentials (overriding the ones
        # configured in the opts / pillar.
        force_reconnect = kwargs.get("force_reconnect", False)
        if force_reconnect:
            log.debug("Usage of reconnect force detected")
            log.debug("Opts before merging")
            log.debug(opts["proxy"])
            opts["proxy"].update(**kwargs)
            log.debug("Opts after merging")
            log.debug(opts["proxy"])
        if is_proxy(opts) and always_alive:
            # if it is running in a NAPALM Proxy and it's using the default
            # always alive behaviour, will get the cached copy of the network
            # device object which should preserve the connection.
            if force_reconnect:
                wrapped_global_namespace["napalm_device"] = get_device(opts)
            else:
                wrapped_global_namespace["napalm_device"] = proxy["napalm.get_device"]()
        elif is_proxy(opts) and not always_alive:
            # if still proxy, but the user does not want the SSH session always alive
            # get a new device instance
            # which establishes a new connection
            # which is closed just before the call() function defined above returns
            if "inherit_napalm_device" not in kwargs or (
                "inherit_napalm_device" in kwargs
                and not kwargs["inherit_napalm_device"]
            ):
                # try to open a new connection
                # but only if the function does not inherit the napalm driver
                # for configuration management this is very important,
                # in order to make sure we are editing the same session.
                try:
                    wrapped_global_namespace["napalm_device"] = get_device(opts)
                except napalm_base.exceptions.ConnectionException as nce:
                    log.error(nce)
                    return "{base_msg}. See log for details.".format(
                        base_msg=str(nce.msg)
                    )
            else:
                # in case the `inherit_napalm_device` is set
                # and it also has a non-empty value,
                # the global var `napalm_device` will be overridden.
                # this is extremely important for configuration-related features
                # as all actions must be issued within the same configuration session
                # otherwise we risk to open multiple sessions
                wrapped_global_namespace["napalm_device"] = kwargs[
                    "inherit_napalm_device"
                ]
        else:
            # if not a NAPLAM proxy
            # thus it is running on a regular minion, directly on the network device
            # or another flavour of Minion from where we can invoke arbitrary
            # NAPALM commands
            # get __salt__ from func_globals
            log.debug("Not running in a NAPALM Proxy Minion")
            _salt_obj = wrapped_global_namespace.get("__salt__")
            napalm_opts = _salt_obj["config.get"]("napalm", {})
            napalm_inventory = _salt_obj["config.get"]("napalm_inventory", {})
            log.debug("NAPALM opts found in the Minion config")
            log.debug(napalm_opts)
            clean_kwargs = salt.utils.args.clean_kwargs(**kwargs)
            napalm_opts.update(clean_kwargs)  # no need for deeper merge
            log.debug("Merging the found opts with the CLI args")
            log.debug(napalm_opts)
            host = (
                napalm_opts.get("host")
                or napalm_opts.get("hostname")
                or napalm_opts.get("fqdn")
                or napalm_opts.get("ip")
            )
            if (
                host
                and napalm_inventory
                and isinstance(napalm_inventory, dict)
                and host in napalm_inventory
            ):
                inventory_opts = napalm_inventory[host]
                log.debug("Found %s in the NAPALM inventory:", host)
                log.debug(inventory_opts)
                napalm_opts.update(inventory_opts)
                log.debug(
                    "Merging the config for %s with the details found in the napalm"
                    " inventory:",
                    host,
                )
                log.debug(napalm_opts)
            opts = copy.deepcopy(opts)  # make sure we don't override the original
            # opts, but just inject the CLI args from the kwargs to into the
            # object manipulated by ``get_device_opts`` to extract the
            # connection details, then use then to establish the connection.
            opts["napalm"] = napalm_opts
            if "inherit_napalm_device" not in kwargs or (
                "inherit_napalm_device" in kwargs
                and not kwargs["inherit_napalm_device"]
            ):
                # try to open a new connection
                # but only if the function does not inherit the napalm driver
                # for configuration management this is very important,
                # in order to make sure we are editing the same session.
                try:
                    wrapped_global_namespace["napalm_device"] = get_device(
                        opts, salt_obj=_salt_obj
                    )
                except napalm_base.exceptions.ConnectionException as nce:
                    log.error(nce)
                    return "{base_msg}. See log for details.".format(
                        base_msg=str(nce.msg)
                    )
            else:
                # in case the `inherit_napalm_device` is set
                # and it also has a non-empty value,
                # the global var `napalm_device` will be overridden.
                # this is extremely important for configuration-related features
                # as all actions must be issued within the same configuration session
                # otherwise we risk to open multiple sessions
                wrapped_global_namespace["napalm_device"] = kwargs[
                    "inherit_napalm_device"
                ]
        if not_always_alive(opts):
            # inject the __opts__ only when not always alive
            # otherwise, we don't want to overload the always-alive proxies
            wrapped_global_namespace["napalm_device"]["__opts__"] = opts
        ret = func(*args, **kwargs)
        if force_reconnect:
            log.debug("That was a forced reconnect, gracefully clearing up")
            device = wrapped_global_namespace["napalm_device"]
            closing = call(device, "close", __retry=False)
        return ret

    return func_wrapper


def default_ret(name):
    """
    Return the default dict of the state output.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    return ret


def loaded_ret(ret, loaded, test, debug, compliance_report=False, opts=None):
    """
    Return the final state output.
    ret
        The initial state output structure.
    loaded
        The loaded dictionary.
    """
    # Always get the comment
    changes = {}
    ret["comment"] = loaded["comment"]
    if "diff" in loaded:
        changes["diff"] = loaded["diff"]
    if "commit_id" in loaded:
        changes["commit_id"] = loaded["commit_id"]
    if "compliance_report" in loaded:
        if compliance_report:
            changes["compliance_report"] = loaded["compliance_report"]
    if debug and "loaded_config" in loaded:
        changes["loaded_config"] = loaded["loaded_config"]
    if changes.get("diff"):
        ret["comment"] = "{comment_base}\n\nConfiguration diff:\n\n{diff}".format(
            comment_base=ret["comment"], diff=changes["diff"]
        )
    if changes.get("loaded_config"):
        ret["comment"] = "{comment_base}\n\nLoaded config:\n\n{loaded_cfg}".format(
            comment_base=ret["comment"], loaded_cfg=changes["loaded_config"]
        )
    if changes.get("compliance_report"):
        ret["comment"] = "{comment_base}\n\nCompliance report:\n\n{compliance}".format(
            comment_base=ret["comment"],
            compliance=salt.output.string_format(
                changes["compliance_report"], "nested", opts=opts
            ),
        )
    if not loaded.get("result", False):
        # Failure of some sort
        return ret
    if not loaded.get("already_configured", True):
        # We're making changes
        if test:
            ret["result"] = None
            return ret
        # Not test, changes were applied
        ret.update(
            {
                "result": True,
                "changes": changes,
                "comment": "Configuration changed!\n{}".format(loaded["comment"]),
            }
        )
        return ret
    # No changes
    ret.update({"result": True, "changes": {}})
    return ret



"""
Apache Libcloud Compute Management
==================================

Connection module for Apache Libcloud Compute management for a full list
of supported clouds, see http://libcloud.readthedocs.io/en/latest/compute/supported_providers.html

Clouds include Amazon EC2, Azure, Google GCE, VMware, OpenStack Nova

.. versionadded:: 2018.3.0

:configuration:
    This module uses a configuration profile for one or multiple cloud providers

    .. code-block:: yaml

        libcloud_compute:
            profile_test1:
              driver: google
              key: service-account@googlecloud.net
              secret: /path/to.key.json
            profile_test2:
              driver: arm
              key: 12345
              secret: mysecret

:depends: apache-libcloud
"""
# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import logging
import os.path

import salt.utils.args
import salt.utils.compat
from salt.utils.versions import LooseVersion as _LooseVersion

log = logging.getLogger(__name__)

REQUIRED_LIBCLOUD_VERSION = "2.0.0"
try:
    # pylint: disable=unused-import
    import libcloud
    from libcloud.compute.providers import get_driver
    from libcloud.compute.base import Node

    # pylint: enable=unused-import
    if hasattr(libcloud, "__version__") and _LooseVersion(
        libcloud.__version__
    ) < _LooseVersion(REQUIRED_LIBCLOUD_VERSION):
        raise ImportError()
    logging.getLogger("libcloud").setLevel(logging.CRITICAL)
    HAS_LIBCLOUD = True
except ImportError:
    HAS_LIBCLOUD = False


def __virtual__():
    """
    Only load if libcloud libraries exist.
    """
    if not HAS_LIBCLOUD:
        return (
            False,
            "A apache-libcloud library with version at least {} was not found".format(
                REQUIRED_LIBCLOUD_VERSION
            ),
        )
    return True


def _get_driver(profile):
    config = __salt__["config.option"]("libcloud_compute")[profile]
    cls = get_driver(config["driver"])
    args = config.copy()
    del args["driver"]
    args["key"] = config.get("key")
    args["secret"] = config.get("secret", None)
    if args["secret"] is None:
        del args["secret"]
    args["secure"] = config.get("secure", True)
    args["host"] = config.get("host", None)
    args["port"] = config.get("port", None)
    return cls(**args)


def list_nodes(profile, **libcloud_kwargs):
    """
    Return a list of nodes

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's list_nodes method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.list_nodes profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    nodes = conn.list_nodes(**libcloud_kwargs)
    ret = []
    for node in nodes:
        ret.append(_simple_node(node))
    return ret


def list_sizes(profile, location_id=None, **libcloud_kwargs):
    """
    Return a list of node sizes

    :param profile: The profile key
    :type  profile: ``str``

    :param location_id: The location key, from list_locations
    :type  location_id: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's list_sizes method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.list_sizes profile1
        salt myminion libcloud_compute.list_sizes profile1 us-east1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    if location_id is not None:
        locations = [loc for loc in conn.list_locations() if loc.id == location_id]
        if not locations:
            raise ValueError("Location not found")
        else:
            sizes = conn.list_sizes(location=locations[0], **libcloud_kwargs)
    else:
        sizes = conn.list_sizes(**libcloud_kwargs)

    ret = []
    for size in sizes:
        ret.append(_simple_size(size))
    return ret


def list_locations(profile, **libcloud_kwargs):
    """
    Return a list of locations for this cloud

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's list_locations method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.list_locations profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    locations = conn.list_locations(**libcloud_kwargs)

    ret = []
    for loc in locations:
        ret.append(_simple_location(loc))
    return ret


def reboot_node(node_id, profile, **libcloud_kwargs):
    """
    Reboot a node in the cloud

    :param node_id: Unique ID of the node to reboot
    :type  node_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's reboot_node method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.reboot_node as-2346 profile1
    """
    conn = _get_driver(profile=profile)
    node = _get_by_id(conn.list_nodes(**libcloud_kwargs), node_id)
    return conn.reboot_node(node, **libcloud_kwargs)


def destroy_node(node_id, profile, **libcloud_kwargs):
    """
    Destroy a node in the cloud

    :param node_id: Unique ID of the node to destroy
    :type  node_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's destroy_node method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.destry_node as-2346 profile1
    """
    conn = _get_driver(profile=profile)
    node = _get_by_id(conn.list_nodes(**libcloud_kwargs), node_id)
    return conn.destroy_node(node, **libcloud_kwargs)


def list_volumes(profile, **libcloud_kwargs):
    """
    Return a list of storage volumes for this cloud

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's list_volumes method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.list_volumes profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    volumes = conn.list_volumes(**libcloud_kwargs)

    ret = []
    for volume in volumes:
        ret.append(_simple_volume(volume))
    return ret


def list_volume_snapshots(volume_id, profile, **libcloud_kwargs):
    """
    Return a list of storage volumes snapshots for this cloud

    :param volume_id: The volume identifier
    :type  volume_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's list_volume_snapshots method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.list_volume_snapshots vol1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    volume = _get_by_id(conn.list_volumes(), volume_id)
    snapshots = conn.list_volume_snapshots(volume, **libcloud_kwargs)

    ret = []
    for snapshot in snapshots:
        ret.append(_simple_volume_snapshot(snapshot))
    return ret


def create_volume(size, name, profile, location_id=None, **libcloud_kwargs):
    """
    Create a storage volume

    :param size: Size of volume in gigabytes (required)
    :type size: ``int``

    :param name: Name of the volume to be created
    :type name: ``str``

    :param location_id: Which data center to create a volume in. If
                            empty, undefined behavior will be selected.
                            (optional)
    :type location_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's list_volumes method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.create_volume 1000 vol1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    if location_id is not None:
        location = _get_by_id(conn.list_locations(), location_id)
    else:
        location = None
    # TODO : Support creating from volume snapshot

    volume = conn.create_volume(size, name, location, snapshot=None, **libcloud_kwargs)
    return _simple_volume(volume)


def create_volume_snapshot(volume_id, profile, name=None, **libcloud_kwargs):
    """
    Create a storage volume snapshot

    :param volume_id:  Volume ID from which to create the new
                        snapshot.
    :type  volume_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param name: Name of the snapshot to be created (optional)
    :type name: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's create_volume_snapshot method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.create_volume_snapshot vol1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    volume = _get_by_id(conn.list_volumes(), volume_id)

    snapshot = conn.create_volume_snapshot(volume, name=name, **libcloud_kwargs)
    return _simple_volume_snapshot(snapshot)


def attach_volume(node_id, volume_id, profile, device=None, **libcloud_kwargs):
    """
    Attaches volume to node.

    :param node_id:  Node ID to target
    :type  node_id: ``str``

    :param volume_id:  Volume ID from which to attach
    :type  volume_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param device: Where the device is exposed, e.g. '/dev/sdb'
    :type device: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's attach_volume method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.detach_volume vol1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    volume = _get_by_id(conn.list_volumes(), volume_id)
    node = _get_by_id(conn.list_nodes(), node_id)
    return conn.attach_volume(node, volume, device=device, **libcloud_kwargs)


def detach_volume(volume_id, profile, **libcloud_kwargs):
    """
    Detaches a volume from a node.

    :param volume_id:  Volume ID from which to detach
    :type  volume_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's detach_volume method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.detach_volume vol1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    volume = _get_by_id(conn.list_volumes(), volume_id)
    return conn.detach_volume(volume, **libcloud_kwargs)


def destroy_volume(volume_id, profile, **libcloud_kwargs):
    """
    Destroy a volume.

    :param volume_id:  Volume ID from which to destroy
    :type  volume_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's destroy_volume method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.destroy_volume vol1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    volume = _get_by_id(conn.list_volumes(), volume_id)
    return conn.destroy_volume(volume, **libcloud_kwargs)


def destroy_volume_snapshot(volume_id, snapshot_id, profile, **libcloud_kwargs):
    """
    Destroy a volume snapshot.

    :param volume_id:  Volume ID from which the snapshot belongs
    :type  volume_id: ``str``

    :param snapshot_id:  Volume Snapshot ID from which to destroy
    :type  snapshot_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's destroy_volume_snapshot method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.destroy_volume_snapshot snap1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    volume = _get_by_id(conn.list_volumes(), volume_id)
    snapshot = _get_by_id(conn.list_volume_snapshots(volume), snapshot_id)
    return conn.destroy_volume_snapshot(snapshot, **libcloud_kwargs)


def list_images(profile, location_id=None, **libcloud_kwargs):
    """
    Return a list of images for this cloud

    :param profile: The profile key
    :type  profile: ``str``

    :param location_id: The location key, from list_locations
    :type  location_id: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's list_images method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.list_images profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    if location_id is not None:
        location = _get_by_id(conn.list_locations(), location_id)
    else:
        location = None
    images = conn.list_images(location=location, **libcloud_kwargs)

    ret = []
    for image in images:
        ret.append(_simple_image(image))
    return ret


def create_image(node_id, name, profile, description=None, **libcloud_kwargs):
    """
    Create an image from a node

    :param node_id: Node to run the task on.
    :type node_id: ``str``

    :param name: name for new image.
    :type name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param description: description for new image.
    :type description: ``description``

    :param libcloud_kwargs: Extra arguments for the driver's create_image method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.create_image server1 my_image profile1
        salt myminion libcloud_compute.create_image server1 my_image profile1 description='test image'
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    node = _get_by_id(conn.list_nodes(), node_id)
    return _simple_image(
        conn.create_image(node, name, description=description, **libcloud_kwargs)
    )


def delete_image(image_id, profile, **libcloud_kwargs):
    """
    Delete an image of a node

    :param image_id: Image to delete
    :type image_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's delete_image method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.delete_image image1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    image = _get_by_id(conn.list_images(), image_id)
    return conn.delete_image(image, **libcloud_kwargs)


def get_image(image_id, profile, **libcloud_kwargs):
    """
    Get an image of a node

    :param image_id: Image to fetch
    :type image_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's delete_image method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.get_image image1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    image = conn.get_image(image_id, **libcloud_kwargs)
    return _simple_image(image)


def copy_image(
    source_region, image_id, name, profile, description=None, **libcloud_kwargs
):
    """
    Copies an image from a source region to the current region.

    :param source_region: Region to copy the node from.
    :type source_region: ``str``

    :param image_id: Image to copy.
    :type image_id: ``str``

    :param name: name for new image.
    :type name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param description: description for new image.
    :type name: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's copy_image method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.copy_image us-east1 image1 'new image' profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    image = conn.get_image(image_id, **libcloud_kwargs)
    new_image = conn.copy_image(
        source_region, image, name, description=description, **libcloud_kwargs
    )
    return _simple_image(new_image)


def list_key_pairs(profile, **libcloud_kwargs):
    """
    List all the available key pair objects.

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's list_key_pairs method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.list_key_pairs profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    keys = conn.list_key_pairs(**libcloud_kwargs)

    ret = []
    for key in keys:
        ret.append(_simple_key_pair(key))
    return ret


def get_key_pair(name, profile, **libcloud_kwargs):
    """
    Get a single key pair by name

    :param name: Name of the key pair to retrieve.
    :type name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's get_key_pair method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.get_key_pair pair1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    return _simple_key_pair(conn.get_key_pair(name, **libcloud_kwargs))


def create_key_pair(name, profile, **libcloud_kwargs):
    """
    Create a single key pair by name

    :param name: Name of the key pair to create.
    :type name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's create_key_pair method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.create_key_pair pair1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    return _simple_key_pair(conn.create_key_pair(name, **libcloud_kwargs))


def import_key_pair(name, key, profile, key_type=None, **libcloud_kwargs):
    """
    Import a new public key from string or a file path

    :param name: Key pair name.
    :type name: ``str``

    :param key: Public key material, the string or a path to a file
    :type  key: ``str`` or path ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param key_type: The key pair type, either `FILE` or `STRING`. Will detect if not provided
        and assume that if the string is a path to an existing path it is a FILE, else STRING.
    :type  key_type: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's import_key_pair_from_xxx method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.import_key_pair pair1 key_value_data123 profile1
        salt myminion libcloud_compute.import_key_pair pair1 /path/to/key profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    if os.path.exists(key) or key_type == "FILE":
        return _simple_key_pair(
            conn.import_key_pair_from_file(name, key, **libcloud_kwargs)
        )
    else:
        return _simple_key_pair(
            conn.import_key_pair_from_string(name, key, **libcloud_kwargs)
        )


def delete_key_pair(name, profile, **libcloud_kwargs):
    """
    Delete a key pair

    :param name: Key pair name.
    :type  name: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's import_key_pair_from_xxx method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.delete_key_pair pair1 profile1
    """
    conn = _get_driver(profile=profile)
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    key = conn.get_key_pair(name)
    return conn.delete_key_pair(key, **libcloud_kwargs)


def extra(method, profile, **libcloud_kwargs):
    """
    Call an extended method on the driver

    :param method: Driver's method name
    :type  method: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param libcloud_kwargs: Extra arguments for the driver's method
    :type  libcloud_kwargs: ``dict``

    CLI Example:

    .. code-block:: bash

        salt myminion libcloud_compute.extra ex_get_permissions google container_name=my_container object_name=me.jpg --out=yaml
    """
    libcloud_kwargs = salt.utils.args.clean_kwargs(**libcloud_kwargs)
    conn = _get_driver(profile=profile)
    connection_method = getattr(conn, method)
    return connection_method(**libcloud_kwargs)


def _get_by_id(collection, id):
    """
    Get item from a list by the id field
    """
    matches = [item for item in collection if item.id == id]
    if not matches:
        raise ValueError("Could not find a matching item")
    elif len(matches) > 1:
        raise ValueError("The id matched {} items, not 1".format(len(matches)))
    return matches[0]


def _simple_volume(volume):
    return {
        "id": volume.id,
        "name": volume.name,
        "size": volume.size,
        "state": volume.state,
        "extra": volume.extra,
    }


def _simple_location(location):
    return {"id": location.id, "name": location.name, "country": location.country}


def _simple_size(size):
    return {
        "id": size.id,
        "name": size.name,
        "ram": size.ram,
        "disk": size.disk,
        "bandwidth": size.bandwidth,
        "price": size.price,
        "extra": size.extra,
    }


def _simple_node(node):
    return {
        "id": node.id,
        "name": node.name,
        "state": str(node.state),
        "public_ips": node.public_ips,
        "private_ips": node.private_ips,
        "size": _simple_size(node.size) if node.size else {},
        "extra": node.extra,
    }


def _simple_volume_snapshot(snapshot):
    return {
        "id": snapshot.id,
        "name": snapshot.name if hasattr(snapshot, "name") else snapshot.id,
        "size": snapshot.size,
        "extra": snapshot.extra,
        "created": snapshot.created,
        "state": snapshot.state,
    }


def _simple_image(image):
    return {
        "id": image.id,
        "name": image.name,
        "extra": image.extra,
    }


def _simple_key_pair(key):
    return {
        "name": key.name,
        "fingerprint": key.fingerprint,
        "public_key": key.public_key,
        "private_key": key.private_key,
        "extra": key.extra,
    }



"""
Management of PostgreSQL languages
==================================

The postgres_language module is used to create and manage Postgres languages.
Languages can be set as either absent or present

.. versionadded:: 2016.3.0

.. code-block:: yaml

    plpgsql:
      postgres_language.present:
        - maintenance_db: testdb

.. code-block:: yaml

    plpgsql:
      postgres_language.absent:
        - maintenance_db: testdb

"""


def __virtual__():
    """
    Only load if the postgres module is present
    """
    if "postgres.language_create" not in __salt__:
        return (
            False,
            "Unable to load postgres module.  Make sure `postgres.bins_dir` is set.",
        )
    return True


def present(
    name,
    maintenance_db,
    user=None,
    db_password=None,
    db_host=None,
    db_port=None,
    db_user=None,
):
    """
    Ensure that a named language is present in the specified
    database.

    name
        The name of the language to install

    maintenance_db
        The name of the database in which the language is to be installed

    user
        System user all operations should be performed on behalf of

    db_user
        database username if different from config or default

    db_password
        user password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Language {} is already installed".format(name),
    }

    dbargs = {
        "runas": user,
        "host": db_host,
        "user": db_user,
        "port": db_port,
        "password": db_password,
    }

    languages = __salt__["postgres.language_list"](maintenance_db, **dbargs)

    if name not in languages:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Language {} is set to be installed".format(name)
            return ret

        if __salt__["postgres.language_create"](name, maintenance_db, **dbargs):
            ret["comment"] = "Language {} has been installed".format(name)
            ret["changes"][name] = "Present"
        else:
            ret["comment"] = "Failed to install language {}".format(name)
            ret["result"] = False

    return ret


def absent(
    name,
    maintenance_db,
    user=None,
    db_password=None,
    db_host=None,
    db_port=None,
    db_user=None,
):
    """
    Ensure that a named language is absent in the specified
    database.

    name
        The name of the language to remove

    maintenance_db
        The name of the database in which the language is to be installed

    user
        System user all operations should be performed on behalf of

    db_user
        database username if different from config or default

    db_password
        user password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    dbargs = {
        "runas": user,
        "host": db_host,
        "user": db_user,
        "port": db_port,
        "password": db_password,
    }

    if __salt__["postgres.language_exists"](name, maintenance_db, **dbargs):
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Language {} is set to be removed".format(name)
            return ret
        if __salt__["postgres.language_remove"](name, **dbargs):
            ret["comment"] = "Language {} has been removed".format(name)
            ret["changes"][name] = "Absent"
            return ret
        else:
            ret["comment"] = "Failed to remove language {}".format(name)
            ret["result"] = False

    ret["comment"] = "Language {} is not present so it cannot be removed".format(name)
    return ret



"""
Test for ssh_pre_flight roster option
"""

import os

import pytest
import salt.utils.files
from tests.support.case import SSHCase
from tests.support.runtests import RUNTIME_VARS


class SSHPreFlightTest(SSHCase):
    """
    Test ssh_pre_flight roster option
    """

    def setUp(self):
        super().setUp()
        self.roster = os.path.join(RUNTIME_VARS.TMP, "pre_flight_roster")
        self.data = {
            "ssh_pre_flight": os.path.join(RUNTIME_VARS.TMP, "ssh_pre_flight.sh")
        }
        self.test_script = os.path.join(
            RUNTIME_VARS.TMP, "test-pre-flight-script-worked.txt"
        )

    def _create_roster(self):
        self.custom_roster(self.roster, self.data)

        with salt.utils.files.fopen(self.data["ssh_pre_flight"], "w") as fp_:
            fp_.write("touch {}".format(self.test_script))

    @pytest.mark.slow_test
    def test_ssh_pre_flight(self):
        """
        test ssh when ssh_pre_flight is set
        ensure the script runs successfully
        """
        self._create_roster()
        assert self.run_function("test.ping", roster_file=self.roster)

        assert os.path.exists(self.test_script)

    @pytest.mark.slow_test
    def test_ssh_run_pre_flight(self):
        """
        test ssh when --pre-flight is passed to salt-ssh
        to ensure the script runs successfully
        """
        self._create_roster()
        # make sure we previously ran a command so the thin dir exists
        self.run_function("test.ping", wipe=False)
        assert not os.path.exists(self.test_script)

        assert self.run_function(
            "test.ping", ssh_opts="--pre-flight", roster_file=self.roster, wipe=False
        )
        assert os.path.exists(self.test_script)

    @pytest.mark.slow_test
    def test_ssh_run_pre_flight_failure(self):
        """
        test ssh_pre_flight when there is a failure
        in the script.
        """
        self._create_roster()
        with salt.utils.files.fopen(self.data["ssh_pre_flight"], "w") as fp_:
            fp_.write("exit 2")

        ret = self.run_function(
            "test.ping", ssh_opts="--pre-flight", roster_file=self.roster, wipe=False
        )
        assert ret["retcode"] == 2

    def tearDown(self):
        """
        make sure to clean up any old ssh directories
        """
        files = [self.roster, self.data["ssh_pre_flight"], self.test_script]
        for fp_ in files:
            if os.path.exists(fp_):
                os.remove(fp_)



import io
import os
import os.path

import attr
import pytest
import salt.config
import salt.loader
from salt.exceptions import SaltRenderError

REQUISITES = ["require", "require_in", "use", "use_in", "watch", "watch_in"]


@attr.s
class Renderer:
    tmp_path = attr.ib()

    def __call__(
        self, content, sls="", saltenv="base", argline="-G yaml . jinja", **kws
    ):
        root_dir = self.tmp_path
        state_tree_dir = self.tmp_path / "state_tree"
        cache_dir = self.tmp_path / "cachedir"
        state_tree_dir.mkdir()
        cache_dir.mkdir()
        config = salt.config.minion_config(None)
        config["root_dir"] = str(root_dir)
        config["state_events"] = False
        config["id"] = "match"
        config["file_client"] = "local"
        config["file_roots"] = dict(base=[str(state_tree_dir)])
        config["cachedir"] = str(cache_dir)
        config["test"] = False
        _renderers = salt.loader.render(config, {"config.get": lambda a, b: False})
        return _renderers["stateconf"](
            io.StringIO(content),
            saltenv=saltenv,
            sls=sls,
            argline=argline,
            renderers=salt.loader.render(config, {}),
            **kws
        )


@pytest.fixture
def renderer(tmp_path):
    return Renderer(tmp_path)


def test_state_config(renderer):
    result = renderer(
        """
.sls_params:
  stateconf.set:
    - name1: value1
    - name2: value2

.extra:
  stateconf:
    - set
    - name: value

# --- end of state config ---

test:
  cmd.run:
    - name: echo name1={{sls_params.name1}} name2={{sls_params.name2}} {{extra.name}}
    - cwd: /
""",
        sls="test",
    )
    assert len(result) == 3
    assert "test::sls_params" in result and "test" in result
    assert "test::extra" in result
    assert (
        result["test"]["cmd.run"][0]["name"] == "echo name1=value1 name2=value2 value"
    )


def test_sls_dir(renderer):
    result = renderer(
        """
test:
  cmd.run:
    - name: echo sls_dir={{sls_dir}}
    - cwd: /
""",
        sls="path.to.sls",
    )
    assert result["test"]["cmd.run"][0]["name"] == "echo sls_dir=path{}to".format(
        os.sep
    )


def test_states_declared_with_shorthand_no_args(renderer):
    result = renderer(
        """
test:
  cmd.run:
    - name: echo testing
    - cwd: /
test1:
  pkg.installed
test2:
  user.present
"""
    )
    assert len(result) == 3
    for args in (result["test1"]["pkg.installed"], result["test2"]["user.present"]):
        assert isinstance(args, list)
        assert len(args) == 0
    assert result["test"]["cmd.run"][0]["name"] == "echo testing"


def test_adding_state_name_arg_for_dot_state_id(renderer):
    result = renderer(
        """
.test:
  pkg.installed:
    - cwd: /
.test2:
  pkg.installed:
    - name: vim
""",
        sls="test",
    )
    assert result["test::test"]["pkg.installed"][0]["name"] == "test"
    assert result["test::test2"]["pkg.installed"][0]["name"] == "vim"


def test_state_prefix(renderer):
    result = renderer(
        """
.test:
  cmd.run:
    - name: echo renamed
    - cwd: /

state_id:
  cmd:
    - run
    - name: echo not renamed
    - cwd: /
""",
        sls="test",
    )
    assert len(result) == 2
    assert "test::test" in result
    assert "state_id" in result


@pytest.mark.parametrize("req", REQUISITES)
def test_dot_state_id_in_requisites(req, renderer):
    result = renderer(
        """
.test:
  cmd.run:
    - name: echo renamed
    - cwd: /

state_id:
  cmd.run:
    - name: echo not renamed
    - cwd: /
    - {}:
      - cmd: .test

""".format(
            req
        ),
        sls="test",
    )
    assert len(result) == 2
    assert "test::test" in result
    assert "state_id" in result
    assert result["state_id"]["cmd.run"][2][req][0]["cmd"] == "test::test"


@pytest.mark.parametrize("req", REQUISITES)
def test_relative_include_with_requisites(req, renderer):
    result = renderer(
        """
include:
  - some.helper
  - .utils

state_id:
  cmd.run:
    - name: echo test
    - cwd: /
    - {}:
      - cmd: .utils::some_state
""".format(
            req
        ),
        sls="test.work",
    )
    assert result["include"][1] == {"base": "test.utils"}
    assert result["state_id"]["cmd.run"][2][req][0]["cmd"] == "test.utils::some_state"


def test_relative_include_and_extend(renderer):
    result = renderer(
        """
include:
  - some.helper
  - .utils

extend:
  .utils::some_state:
    cmd.run:
      - name: echo overridden
    """,
        sls="test.work",
    )
    assert "test.utils::some_state" in result["extend"]


@pytest.mark.parametrize("req", REQUISITES)
def test_multilevel_relative_include_with_requisites(req, renderer):
    result = renderer(
        """
include:
  - .shared
  - ..utils
  - ...helper

state_id:
  cmd.run:
    - name: echo test
    - cwd: /
    - {}:
      - cmd: ..utils::some_state
""".format(
            req
        ),
        sls="test.nested.work",
    )
    assert result["include"][0] == {"base": "test.nested.shared"}
    assert result["include"][1] == {"base": "test.utils"}
    assert result["include"][2] == {"base": "helper"}
    assert result["state_id"]["cmd.run"][2][req][0]["cmd"] == "test.utils::some_state"


def test_multilevel_relative_include_beyond_top_level(renderer):
    pytest.raises(
        SaltRenderError,
        renderer,
        """
include:
  - ...shared
""",
        sls="test.work",
    )


def test_start_state_generation(renderer):
    result = renderer(
        """
A:
  cmd.run:
    - name: echo hello
    - cwd: /
B:
  cmd.run:
    - name: echo world
    - cwd: /
""",
        sls="test",
        argline="-so yaml . jinja",
    )
    assert len(result) == 4
    assert result["test::start"]["stateconf.set"][0]["require_in"][0]["cmd"] == "A"


def test_goal_state_generation(renderer):
    result = renderer(
        """
{% for sid in "ABCDE": %}
{{sid}}:
  cmd.run:
    - name: echo this is {{sid}}
    - cwd: /
{% endfor %}

""",
        sls="test.goalstate",
        argline="yaml . jinja",
    )
    assert len(result) == len("ABCDE") + 1

    reqs = result["test.goalstate::goal"]["stateconf.set"][0]["require"]
    assert {next(iter(i.values())) for i in reqs} == set("ABCDE")


def test_implicit_require_with_goal_state(renderer):
    result = renderer(
        """
{% for sid in "ABCDE": %}
{{sid}}:
  cmd.run:
    - name: echo this is {{sid}}
    - cwd: /
{% endfor %}

F:
  cmd.run:
    - name: echo this is F
    - cwd: /
    - require:
      - cmd: A
      - cmd: B

G:
  cmd.run:
    - name: echo this is G
    - cwd: /
    - require:
      - cmd: D
      - cmd: F
""",
        sls="test",
        argline="-o yaml . jinja",
    )

    sids = "ABCDEFG"[::-1]
    for i, sid in enumerate(sids):
        if i < len(sids) - 1:
            assert result[sid]["cmd.run"][2]["require"][0]["cmd"] == sids[i + 1]

    F_args = result["F"]["cmd.run"]
    assert len(F_args) == 3
    F_req = F_args[2]["require"]
    assert len(F_req) == 3
    assert F_req[1]["cmd"] == "A"
    assert F_req[2]["cmd"] == "B"

    G_args = result["G"]["cmd.run"]
    assert len(G_args) == 3
    G_req = G_args[2]["require"]
    assert len(G_req) == 3
    assert G_req[1]["cmd"] == "D"
    assert G_req[2]["cmd"] == "F"

    goal_args = result["test::goal"]["stateconf.set"]
    assert len(goal_args) == 1
    assert [next(iter(i.values())) for i in goal_args[0]["require"]] == list("ABCDEFG")


def test_slsdir(renderer):
    result = renderer(
        """
formula/woot.sls:
  cmd.run:
    - name: echo {{ slspath }}
    - cwd: /
""",
        sls="formula.woot",
        argline="yaml . jinja",
    )

    r = result["formula/woot.sls"]["cmd.run"][0]["name"]
    assert r == "echo formula/woot"



import pytest
import salt.utils.platform
import salt.utils.win_lgpo_netsh as win_lgpo_netsh
from salt.exceptions import CommandExecutionError
from tests.support.mock import patch
from tests.support.unit import TestCase, skipIf


@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
class WinLgpoNetshTestCase(TestCase):
    def test_get_settings_firewallpolicy_local(self):
        ret = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store="local"
        )
        self.assertIn("Inbound", ret)
        self.assertIn("Outbound", ret)

    def test_get_settings_firewallpolicy_lgpo(self):
        ret = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store="lgpo"
        )
        self.assertIn("Inbound", ret)
        self.assertIn("Outbound", ret)

    def test_get_settings_firewallpolicy_lgpo_issue_57591(self):
        """
        Should not stacktrace when the hostname contains unicode characters
        """
        with patch.object(win_lgpo_netsh, "__hostname__", return_value="kомпьютер"):
            ret = win_lgpo_netsh.get_settings(
                profile="domain", section="firewallpolicy", store="lgpo"
            )
        self.assertIn("Inbound", ret)
        self.assertIn("Outbound", ret)

    def test_get_settings_logging_local(self):
        ret = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="local"
        )
        self.assertIn("FileName", ret)
        self.assertIn("LogAllowedConnections", ret)
        self.assertIn("LogDroppedConnections", ret)
        self.assertIn("MaxFileSize", ret)

    def test_get_settings_logging_lgpo(self):
        ret = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="lgpo"
        )
        self.assertIn("FileName", ret)
        self.assertIn("LogAllowedConnections", ret)
        self.assertIn("LogDroppedConnections", ret)
        self.assertIn("MaxFileSize", ret)

    def test_get_settings_settings_local(self):
        ret = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="local"
        )
        self.assertIn("InboundUserNotification", ret)
        self.assertIn("LocalConSecRules", ret)
        self.assertIn("LocalFirewallRules", ret)
        self.assertIn("RemoteManagement", ret)
        self.assertIn("UnicastResponseToMulticast", ret)

    def test_get_settings_settings_lgpo(self):
        ret = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="lgpo"
        )
        self.assertIn("InboundUserNotification", ret)
        self.assertIn("LocalConSecRules", ret)
        self.assertIn("LocalFirewallRules", ret)
        self.assertIn("RemoteManagement", ret)
        self.assertIn("UnicastResponseToMulticast", ret)

    def test_get_settings_state_local(self):
        ret = win_lgpo_netsh.get_settings(
            profile="domain", section="state", store="local"
        )
        self.assertIn("State", ret)

    def test_get_settings_state_lgpo(self):
        ret = win_lgpo_netsh.get_settings(
            profile="domain", section="state", store="lgpo"
        )
        self.assertIn("State", ret)

    def test_get_all_settings_local(self):
        ret = win_lgpo_netsh.get_all_settings(profile="domain", store="local")

        self.assertIn("Inbound", ret)
        self.assertIn("Outbound", ret)
        self.assertIn("FileName", ret)
        self.assertIn("LogAllowedConnections", ret)
        self.assertIn("LogDroppedConnections", ret)
        self.assertIn("MaxFileSize", ret)
        self.assertIn("InboundUserNotification", ret)
        self.assertIn("LocalConSecRules", ret)
        self.assertIn("LocalFirewallRules", ret)
        self.assertIn("RemoteManagement", ret)
        self.assertIn("UnicastResponseToMulticast", ret)
        self.assertIn("State", ret)

    def test_get_all_settings_lgpo(self):
        ret = win_lgpo_netsh.get_all_settings(profile="domain", store="local")

        self.assertIn("Inbound", ret)
        self.assertIn("Outbound", ret)
        self.assertIn("FileName", ret)
        self.assertIn("LogAllowedConnections", ret)
        self.assertIn("LogDroppedConnections", ret)
        self.assertIn("MaxFileSize", ret)
        self.assertIn("InboundUserNotification", ret)
        self.assertIn("LocalConSecRules", ret)
        self.assertIn("LocalFirewallRules", ret)
        self.assertIn("RemoteManagement", ret)
        self.assertIn("UnicastResponseToMulticast", ret)
        self.assertIn("State", ret)

    def test_get_all_profiles_local(self):
        ret = win_lgpo_netsh.get_all_profiles(store="local")
        self.assertIn("Domain Profile", ret)
        self.assertIn("Private Profile", ret)
        self.assertIn("Public Profile", ret)

    def test_get_all_profiles_lgpo(self):
        ret = win_lgpo_netsh.get_all_profiles(store="lgpo")
        self.assertIn("Domain Profile", ret)
        self.assertIn("Private Profile", ret)
        self.assertIn("Public Profile", ret)

    @pytest.mark.destructive_test
    def test_set_firewall_settings_inbound_local(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store="local"
        )["Inbound"]
        try:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", inbound="allowinbound", store="local"
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="firewallpolicy", store="local"
            )["Inbound"]
            self.assertEqual("AllowInbound", new)
        finally:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", inbound=current, store="local"
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_settings_inbound_local_notconfigured(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store="local"
        )["Inbound"]
        try:
            self.assertRaises(
                CommandExecutionError,
                win_lgpo_netsh.set_firewall_settings,
                profile="domain",
                inbound="notconfigured",
                store="local",
            )
        finally:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", inbound=current, store="local"
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_settings_inbound_lgpo_notconfigured(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store="lgpo"
        )["Inbound"]
        try:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", inbound="notconfigured", store="lgpo"
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="firewallpolicy", store="lgpo"
            )["Inbound"]
            self.assertEqual("NotConfigured", new)
        finally:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", inbound=current, store="lgpo"
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_settings_outbound_local(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store="local"
        )["Outbound"]
        try:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", outbound="allowoutbound", store="local"
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="firewallpolicy", store="local"
            )["Outbound"]
            self.assertEqual("AllowOutbound", new)
        finally:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", outbound=current, store="local"
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_logging_allowed_local_enable(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="local"
        )["LogAllowedConnections"]
        try:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain",
                setting="allowedconnections",
                value="enable",
                store="local",
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="logging", store="local"
            )["LogAllowedConnections"]
            self.assertEqual("Enable", new)
        finally:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain",
                setting="allowedconnections",
                value=current,
                store="local",
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_logging_allowed_local_notconfigured(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="local"
        )["LogAllowedConnections"]
        try:
            self.assertRaises(
                CommandExecutionError,
                win_lgpo_netsh.set_logging_settings,
                profile="domain",
                setting="allowedconnections",
                value="notconfigured",
                store="local",
            )
        finally:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain",
                setting="allowedconnections",
                value=current,
                store="local",
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_logging_allowed_lgpo_notconfigured(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="lgpo"
        )["LogAllowedConnections"]
        try:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain",
                setting="allowedconnections",
                value="notconfigured",
                store="lgpo",
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="logging", store="lgpo"
            )["LogAllowedConnections"]
            self.assertEqual("NotConfigured", new)
        finally:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain",
                setting="allowedconnections",
                value=current,
                store="lgpo",
            )
            self.assertTrue(ret)

    def test_set_firewall_logging_dropped_local_enable(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="local"
        )["LogDroppedConnections"]
        try:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain",
                setting="droppedconnections",
                value="enable",
                store="local",
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="logging", store="local"
            )["LogDroppedConnections"]
            self.assertEqual("Enable", new)
        finally:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain",
                setting="droppedconnections",
                value=current,
                store="local",
            )
            self.assertTrue(ret)

    def test_set_firewall_logging_filename_local(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="local"
        )["FileName"]
        try:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain",
                setting="filename",
                value="C:\\Temp\\test.log",
                store="local",
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="logging", store="local"
            )["FileName"]
            self.assertEqual("C:\\Temp\\test.log", new)
        finally:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain", setting="filename", value=current, store="local"
            )
            self.assertTrue(ret)

    def test_set_firewall_logging_maxfilesize_local(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="local"
        )["MaxFileSize"]
        try:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain", setting="maxfilesize", value="16384", store="local"
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="logging", store="local"
            )["MaxFileSize"]
            self.assertEqual("16384", new)
        finally:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain", setting="maxfilesize", value=current, store="local"
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_settings_fwrules_local_enable(self):
        self.assertRaises(
            CommandExecutionError,
            win_lgpo_netsh.set_settings,
            profile="domain",
            setting="localfirewallrules",
            value="enable",
            store="local",
        )

    @pytest.mark.destructive_test
    def test_set_firewall_settings_fwrules_lgpo_notconfigured(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="lgpo"
        )["LocalFirewallRules"]
        try:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="localfirewallrules",
                value="notconfigured",
                store="lgpo",
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="settings", store="lgpo"
            )["LocalFirewallRules"]
            self.assertEqual("NotConfigured", new)
        finally:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="localfirewallrules",
                value=current,
                store="lgpo",
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_settings_consecrules_local_enable(self):
        self.assertRaises(
            CommandExecutionError,
            win_lgpo_netsh.set_settings,
            profile="domain",
            setting="localconsecrules",
            value="enable",
            store="local",
        )

    def test_set_firewall_settings_notification_local_enable(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="local"
        )["InboundUserNotification"]
        try:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="inboundusernotification",
                value="enable",
                store="local",
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="settings", store="local"
            )["InboundUserNotification"]
            self.assertEqual("Enable", new)
        finally:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="inboundusernotification",
                value=current,
                store="local",
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_settings_notification_local_notconfigured(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="local"
        )["InboundUserNotification"]
        try:
            self.assertRaises(
                CommandExecutionError,
                win_lgpo_netsh.set_settings,
                profile="domain",
                setting="inboundusernotification",
                value="notconfigured",
                store="local",
            )
        finally:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="inboundusernotification",
                value=current,
                store="local",
            )
            self.assertTrue(ret)

    def test_set_firewall_settings_notification_lgpo_notconfigured(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="lgpo"
        )["InboundUserNotification"]
        try:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="inboundusernotification",
                value="notconfigured",
                store="lgpo",
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="settings", store="lgpo"
            )["InboundUserNotification"]
            self.assertEqual("NotConfigured", new)
        finally:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="inboundusernotification",
                value=current,
                store="lgpo",
            )
            self.assertTrue(ret)

    def test_set_firewall_settings_remotemgmt_local_enable(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="local"
        )["RemoteManagement"]
        try:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="remotemanagement",
                value="enable",
                store="local",
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="settings", store="local"
            )["RemoteManagement"]
            self.assertEqual("Enable", new)
        finally:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="remotemanagement",
                value=current,
                store="local",
            )
            self.assertTrue(ret)

    def test_set_firewall_settings_unicast_local_disable(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="local"
        )["UnicastResponseToMulticast"]
        try:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="unicastresponsetomulticast",
                value="disable",
                store="local",
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="settings", store="local"
            )["UnicastResponseToMulticast"]
            self.assertEqual("Disable", new)
        finally:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting="unicastresponsetomulticast",
                value=current,
                store="local",
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_state_local_on(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="state", store="local"
        )["State"]
        try:
            ret = win_lgpo_netsh.set_state(profile="domain", state="off", store="local")
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="state", store="local"
            )["State"]
            self.assertEqual("OFF", new)
        finally:
            ret = win_lgpo_netsh.set_state(
                profile="domain", state=current, store="local"
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_state_local_notconfigured(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="state", store="local"
        )["State"]
        try:
            self.assertRaises(
                CommandExecutionError,
                win_lgpo_netsh.set_state,
                profile="domain",
                state="notconfigured",
                store="local",
            )
        finally:
            ret = win_lgpo_netsh.set_state(
                profile="domain", state=current, store="local"
            )
            self.assertTrue(ret)

    @pytest.mark.destructive_test
    def test_set_firewall_state_lgpo_notconfigured(self):
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="state", store="local"
        )["State"]
        try:
            ret = win_lgpo_netsh.set_state(
                profile="domain", state="notconfigured", store="lgpo"
            )
            self.assertTrue(ret)
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="state", store="lgpo"
            )["State"]
            self.assertEqual("NotConfigured", new)
        finally:
            ret = win_lgpo_netsh.set_state(
                profile="domain", state=current, store="lgpo"
            )
            self.assertTrue(ret)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.sqlite3 as sqlite3
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase


class MockSqlite3:
    """
    Mock sqlite3 class
    """

    version = "2.6.0"
    sqlite_version = "3.8.2"

    def __init__(self):
        self.dbase = None
        self.isolation_level = None

    def connect(self, dbase, isolation_level=None):
        """
        Mock connect method
        """
        self.dbase = dbase
        self.isolation_level = isolation_level
        return MockSqlite3()

    @staticmethod
    def cursor():
        """
        Mock connect method
        """
        return MockSqlite3()

    @staticmethod
    def execute(sql):
        """
        Mock connect method
        """
        return sql

    @staticmethod
    def fetchall():
        """
        Mock connect method
        """
        return True


class Sqlite3TestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.sqlite3
    """

    def setup_loader_modules(self):
        return {sqlite3: {"sqlite3": MockSqlite3()}}

    # 'version' function tests: 1

    def test_version(self):
        """
        Tests if it return version of pysqlite.
        """
        self.assertEqual(sqlite3.version(), "2.6.0")

    # 'sqlite_version' function tests: 1

    def test_sqlite_version(self):
        """
        Tests if it return version of sqlite.
        """
        self.assertEqual(sqlite3.sqlite_version(), "3.8.2")

    # 'modify' function tests: 1

    def test_modify(self):
        """
        Tests if it issue an SQL query to sqlite3 (with no return data).
        """
        self.assertFalse(sqlite3.modify())

        self.assertTrue(
            sqlite3.modify("/root/test.db", "CREATE TABLE test(id INT, testdata TEXT);")
        )

    # 'fetch' function tests: 1

    def test_fetch(self):
        """
        Tests if it retrieve data from an sqlite3 db
        (returns all rows, be careful!)
        """
        self.assertFalse(sqlite3.fetch())

        self.assertTrue(
            sqlite3.fetch("/root/test.db", "CREATE TABLE test(id INT, testdata TEXT);")
        )

    # 'tables' function tests: 1

    def test_tables(self):
        """
        Tests if it show all tables in the database.
        """
        self.assertFalse(sqlite3.tables())

        self.assertTrue(sqlite3.tables("/root/test.db"))

    # 'indices' function tests: 1

    def test_indices(self):
        """
        Tests if it show all indices in the database.
        """
        self.assertFalse(sqlite3.indices())

        self.assertTrue(sqlite3.indices("/root/test.db"))

    # 'indexes' function tests: 1

    def test_indexes(self):
        """
        Tests if it show all indices in the database,
        for people with poor spelling skills
        """
        self.assertTrue(sqlite3.indexes("/root/test.db"))



import pytest
import salt.states.postgres_schema as postgres_schema
from tests.support.mock import MagicMock, Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        postgres_schema: {
            "__grains__": {"os_family": "Linux"},
            "__salt__": {
                "config.option": Mock(),
                "cmd.run_all": Mock(),
                "file.chown": Mock(),
                "file.remove": Mock(),
            },
            "__opts__": {"test": False},
        },
    }


def test_present_creation():
    with patch.dict(
        postgres_schema.__salt__,
        {
            "postgres.schema_get": Mock(return_value=None),
            "postgres.schema_create": MagicMock(),
        },
    ):
        ret = postgres_schema.present("dbname", "foo")
        assert ret == {
            "comment": "Schema foo has been created in database dbname",
            "changes": {"foo": "Present"},
            "dbname": "dbname",
            "name": "foo",
            "result": True,
        }
        assert postgres_schema.__salt__["postgres.schema_create"].call_count == 1


def test_present_nocreation():
    with patch.dict(
        postgres_schema.__salt__,
        {
            "postgres.schema_get": Mock(
                return_value={"foo": {"acl": "", "owner": "postgres"}}
            ),
            "postgres.schema_create": MagicMock(),
        },
    ):
        ret = postgres_schema.present("dbname", "foo")
        assert ret == {
            "comment": "Schema foo already exists in database dbname",
            "changes": {},
            "dbname": "dbname",
            "name": "foo",
            "result": True,
        }
        assert postgres_schema.__salt__["postgres.schema_create"].call_count == 0


def test_absent_remove():
    with patch.dict(
        postgres_schema.__salt__,
        {
            "postgres.schema_exists": Mock(return_value=True),
            "postgres.schema_remove": MagicMock(),
        },
    ):
        ret = postgres_schema.absent("dbname", "foo")
        assert ret == {
            "comment": "Schema foo has been removed from database dbname",
            "changes": {"foo": "Absent"},
            "dbname": "dbname",
            "name": "foo",
            "result": True,
        }
        assert postgres_schema.__salt__["postgres.schema_remove"].call_count == 1


def test_absent_noremove():
    with patch.dict(
        postgres_schema.__salt__,
        {
            "postgres.schema_exists": Mock(return_value=False),
            "postgres.schema_remove": MagicMock(),
        },
    ):
        ret = postgres_schema.absent("dbname", "foo")
        assert ret == {
            "comment": (
                "Schema foo is not present in database dbname, so it cannot be removed"
            ),
            "changes": {},
            "dbname": "dbname",
            "name": "foo",
            "result": True,
        }
        assert postgres_schema.__salt__["postgres.schema_remove"].call_count == 0



import salt.modules.saltutil as saltutil
from salt.client import LocalClient
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import create_autospec
from tests.support.mock import sentinel as s
from tests.support.unit import TestCase


class ScheduleTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {saltutil: {}}

    def test_exec_kwargs(self):
        _cmd_expected_kwargs = {
            "tgt": s.tgt,
            "fun": s.fun,
            "arg": s.arg,
            "timeout": s.timeout,
            "tgt_type": s.tgt_type,
            "ret": s.ret,
            "kwarg": s.kwarg,
        }
        client = create_autospec(LocalClient)

        saltutil._exec(client, **_cmd_expected_kwargs)
        client.cmd_iter.assert_called_with(**_cmd_expected_kwargs)

        saltutil._exec(
            client,
            s.tgt,
            s.fun,
            s.arg,
            s.timeout,
            s.tgt_type,
            s.ret,
            s.kwarg,
            **{"batch": s.batch}
        )
        client.cmd_batch.assert_called_with(batch=s.batch, **_cmd_expected_kwargs)

        saltutil._exec(
            client,
            s.tgt,
            s.fun,
            s.arg,
            s.timeout,
            s.tgt_type,
            s.ret,
            s.kwarg,
            **{"subset": s.subset}
        )
        client.cmd_subset.assert_called_with(
            subset=s.subset, cli=True, **_cmd_expected_kwargs
        )

        saltutil._exec(
            client,
            s.tgt,
            s.fun,
            s.arg,
            s.timeout,
            s.tgt_type,
            s.ret,
            s.kwarg,
            **{"subset": s.subset, "cli": s.cli}
        )
        client.cmd_subset.assert_called_with(
            subset=s.subset, cli=s.cli, **_cmd_expected_kwargs
        )

        # cmd_batch doesn't know what to do with 'subset', don't pass it along.
        saltutil._exec(
            client,
            s.tgt,
            s.fun,
            s.arg,
            s.timeout,
            s.tgt_type,
            s.ret,
            s.kwarg,
            **{"subset": s.subset, "batch": s.batch}
        )
        client.cmd_batch.assert_called_with(batch=s.batch, **_cmd_expected_kwargs)



import os
import shutil

import pytest
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.unit import skipIf


@pytest.mark.windows_whitelisted
@skipIf(salt.utils.platform.is_darwin(), "No mtab on Darwin")
@skipIf(salt.utils.platform.is_freebsd(), "No mtab on FreeBSD")
@skipIf(salt.utils.platform.is_windows(), "No mtab on Windows")
@pytest.mark.destructive_test
class DiskModuleVirtualizationTest(ModuleCase):
    """
    Test to make sure we return a clean result under Docker. Refs #8976

    This is factored into its own class so that we can have some certainty that setUp() and tearDown() are run.
    """

    def setUp(self):
        # Make /etc/mtab unreadable
        if os.path.isfile("/etc/mtab"):
            shutil.move("/etc/mtab", "/tmp/mtab")

    def test_no_mtab(self):
        ret = self.run_function("disk.usage")
        self.assertDictEqual(ret, {})

    def tearDown(self):
        if os.path.isfile("/tmp/mtab"):
            shutil.move("/tmp/mtab", "/etc/mtab")


@pytest.mark.windows_whitelisted
class DiskModuleTest(ModuleCase):
    """
    Validate the disk module
    """

    @pytest.mark.slow_test
    def test_usage(self):
        """
        disk.usage
        """
        ret = self.run_function("disk.usage")
        self.assertTrue(isinstance(ret, dict))
        if not isinstance(ret, dict):
            return
        if salt.utils.platform.is_darwin():
            for key, val in ret.items():
                self.assertTrue("filesystem" in val)
                self.assertTrue("512-blocks" in val)
                self.assertTrue("used" in val)
                self.assertTrue("available" in val)
                self.assertTrue("capacity" in val)
                self.assertTrue("iused" in val)
                self.assertTrue("ifree" in val)
                self.assertTrue("%iused" in val)
        else:
            for key, val in ret.items():
                self.assertTrue("filesystem" in val)
                self.assertTrue("1K-blocks" in val)
                self.assertTrue("used" in val)
                self.assertTrue("available" in val)
                self.assertTrue("capacity" in val)

    @skipIf(salt.utils.platform.is_windows(), "inode info not available on Windows")
    def test_inodeusage(self):
        """
        disk.inodeusage
        """
        ret = self.run_function("disk.inodeusage")
        self.assertTrue(isinstance(ret, dict))
        if not isinstance(ret, dict):
            return
        for key, val in ret.items():
            self.assertTrue("inodes" in val)
            self.assertTrue("used" in val)
            self.assertTrue("free" in val)
            self.assertTrue("use" in val)
            self.assertTrue("filesystem" in val)



"""
Test module for syslog_ng
"""

# Import Python modules

import os
from textwrap import dedent

import salt.modules.syslog_ng as syslog_ng
import salt.utils.platform
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

_VERSION = "3.6.0alpha0"
_MODULES = (
    "syslogformat,json-plugin,basicfuncs,afstomp,afsocket,cryptofuncs,"
    "afmongodb,dbparser,system-source,affile,pseudofile,afamqp,"
    "afsocket-notls,csvparser,linux-kmsg-format,afuser,confgen,afprog"
)

VERSION_OUTPUT = """syslog-ng {0}
Installer-Version: {0}
Revision:
Compile-Date: Apr  4 2014 20:26:18
Error opening plugin module; module='afsocket-tls', error='/home/tibi/install/syslog-ng/lib/syslog-ng/libafsocket-tls.so: undefined symbol: tls_context_setup_session'
Available-Modules: {1}
Enable-Debug: on
Enable-GProf: off
Enable-Memtrace: off
Enable-IPv6: on
Enable-Spoof-Source: off
Enable-TCP-Wrapper: off
Enable-Linux-Caps: off""".format(
    _VERSION, _MODULES
)

STATS_OUTPUT = """SourceName;SourceId;SourceInstance;State;Type;Number
center;;received;a;processed;0
destination;#anon-destination0;;a;processed;0
destination;#anon-destination1;;a;processed;0
source;s_gsoc2014;;a;processed;0
center;;queued;a;processed;0
global;payload_reallocs;;a;processed;0
global;sdata_updates;;a;processed;0
global;msg_clones;;a;processed;0"""

_SYSLOG_NG_NOT_INSTALLED_RETURN_VALUE = {
    "retcode": -1,
    "stderr": "Unable to execute the command 'syslog-ng'. It is not in the PATH.",
}
_SYSLOG_NG_CTL_NOT_INSTALLED_RETURN_VALUE = {
    "retcode": -1,
    "stderr": "Unable to execute the command 'syslog-ng-ctl'. It is not in the PATH.",
}


class SyslogNGTestCase(TestCase, LoaderModuleMockMixin):

    # pylint: disable=blacklisted-function
    orig_env = {"PATH": "/foo:/bar"}
    bin_dir = "/baz"
    mocked_env = {"PATH": "/foo:/bar:/baz"}
    # pylint: enable=blacklisted-function

    def setup_loader_modules(self):
        return {syslog_ng: {}}

    def test_statement_without_options(self):
        s = syslog_ng.Statement("source", "s_local", options=[])
        b = s.build()
        self.assertEqual(
            dedent(
                """\
            source s_local {
            };
            """
            ),
            b,
        )

    def test_non_empty_statement(self):
        o1 = syslog_ng.Option("file")
        o2 = syslog_ng.Option("tcp")
        s = syslog_ng.Statement("source", "s_local", options=[o1, o2])
        b = s.build()
        self.assertEqual(
            dedent(
                """\
            source s_local {
                file(
                );
                tcp(
                );
            };
            """
            ),
            b,
        )

    def test_option_with_parameters(self):
        o1 = syslog_ng.Option("file")
        p1 = syslog_ng.SimpleParameter('"/var/log/messages"')
        p2 = syslog_ng.SimpleParameter()
        p3 = syslog_ng.TypedParameter()
        p3.type = "tls"
        p2.value = '"/var/log/syslog"'
        o1.add_parameter(p1)
        o1.add_parameter(p2)
        o1.add_parameter(p3)
        b = o1.build()
        self.assertEqual(
            dedent(
                """\
            file(
                "/var/log/messages",
                "/var/log/syslog",
                tls(
                )
            );
            """
            ),
            b,
        )

    def test_parameter_with_values(self):
        p = syslog_ng.TypedParameter()
        p.type = "tls"
        v1 = syslog_ng.TypedParameterValue()
        v1.type = "key_file"

        v2 = syslog_ng.TypedParameterValue()
        v2.type = "cert_file"

        p.add_value(v1)
        p.add_value(v2)

        b = p.build()
        self.assertEqual(
            dedent(
                """\
            tls(
                key_file(
                ),
                cert_file(
                )
            )"""
            ),
            b,
        )

    def test_value_with_arguments(self):
        t = syslog_ng.TypedParameterValue()
        t.type = "key_file"

        a1 = syslog_ng.Argument('"/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"')
        a2 = syslog_ng.Argument('"/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"')

        t.add_argument(a1)
        t.add_argument(a2)

        b = t.build()
        self.assertEqual(
            dedent(
                """\
            key_file(
                "/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"
                "/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"
            )"""
            ),
            b,
        )

    def test_end_to_end_statement_generation(self):
        s = syslog_ng.Statement("source", "s_tls")

        o = syslog_ng.Option("tcp")

        ip = syslog_ng.TypedParameter("ip")
        ip.add_value(syslog_ng.SimpleParameterValue("'192.168.42.2'"))
        o.add_parameter(ip)

        port = syslog_ng.TypedParameter("port")
        port.add_value(syslog_ng.SimpleParameterValue(514))
        o.add_parameter(port)

        tls = syslog_ng.TypedParameter("tls")
        key_file = syslog_ng.TypedParameterValue("key_file")
        key_file.add_argument(
            syslog_ng.Argument('"/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"')
        )
        cert_file = syslog_ng.TypedParameterValue("cert_file")
        cert_file.add_argument(
            syslog_ng.Argument('"/opt/syslog-ng/etc/syslog-ng/cert.d/syslog-ng.cert"')
        )
        peer_verify = syslog_ng.TypedParameterValue("peer_verify")
        peer_verify.add_argument(syslog_ng.Argument("optional-untrusted"))
        tls.add_value(key_file)
        tls.add_value(cert_file)
        tls.add_value(peer_verify)
        o.add_parameter(tls)

        s.add_child(o)
        b = s.build()
        self.assertEqual(
            dedent(
                """\
            source s_tls {
                tcp(
                    ip(
                        '192.168.42.2'
                    ),
                    port(
                        514
                    ),
                    tls(
                        key_file(
                            "/opt/syslog-ng/etc/syslog-ng/key.d/syslog-ng.key"
                        ),
                        cert_file(
                            "/opt/syslog-ng/etc/syslog-ng/cert.d/syslog-ng.cert"
                        ),
                        peer_verify(
                            optional-untrusted
                        )
                    )
                );
            };
            """
            ),
            b,
        )

    @skipIf(salt.utils.platform.is_windows(), "Module not available on Windows")
    def test_version(self):
        cmd_ret = {"retcode": 0, "stdout": VERSION_OUTPUT}
        expected_output = {"retcode": 0, "stdout": _VERSION}
        cmd_args = ["syslog-ng", "-V"]

        cmd_mock = MagicMock(return_value=cmd_ret)
        with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
            os.environ, self.orig_env
        ):
            result = syslog_ng.version()
            self.assertEqual(result, expected_output)
            cmd_mock.assert_called_once_with(cmd_args, env=None, python_shell=False)

        cmd_mock = MagicMock(return_value=cmd_ret)
        with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
            os.environ, self.orig_env
        ):
            result = syslog_ng.version(syslog_ng_sbin_dir=self.bin_dir)
            self.assertEqual(result, expected_output)
            cmd_mock.assert_called_once_with(
                cmd_args, env=self.mocked_env, python_shell=False
            )

    @skipIf(salt.utils.platform.is_windows(), "Module not available on Windows")
    def test_stats(self):
        cmd_ret = {"retcode": 0, "stdout": STATS_OUTPUT}
        cmd_args = ["syslog-ng-ctl", "stats"]

        cmd_mock = MagicMock(return_value=cmd_ret)
        with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
            os.environ, self.orig_env
        ):
            result = syslog_ng.stats()
            self.assertEqual(result, cmd_ret)
            cmd_mock.assert_called_once_with(cmd_args, env=None, python_shell=False)

        cmd_mock = MagicMock(return_value=cmd_ret)
        with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
            os.environ, self.orig_env
        ):
            result = syslog_ng.stats(syslog_ng_sbin_dir=self.bin_dir)
            self.assertEqual(result, cmd_ret)
            cmd_mock.assert_called_once_with(
                cmd_args, env=self.mocked_env, python_shell=False
            )

    @skipIf(salt.utils.platform.is_windows(), "Module not available on Windows")
    def test_modules(self):
        cmd_ret = {"retcode": 0, "stdout": VERSION_OUTPUT}
        expected_output = {"retcode": 0, "stdout": _MODULES}
        cmd_args = ["syslog-ng", "-V"]

        cmd_mock = MagicMock(return_value=cmd_ret)
        with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
            os.environ, self.orig_env
        ):
            result = syslog_ng.modules()
            self.assertEqual(result, expected_output)
            cmd_mock.assert_called_once_with(cmd_args, env=None, python_shell=False)

        cmd_mock = MagicMock(return_value=cmd_ret)
        with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
            os.environ, self.orig_env
        ):
            result = syslog_ng.modules(syslog_ng_sbin_dir=self.bin_dir)
            self.assertEqual(result, expected_output)
            cmd_mock.assert_called_once_with(
                cmd_args, env=self.mocked_env, python_shell=False
            )

    @skipIf(salt.utils.platform.is_windows(), "Module not available on Windows")
    def test_config_test(self):
        cmd_ret = {"retcode": 0, "stderr": "", "stdout": "Foo"}
        cmd_args = ["syslog-ng", "--syntax-only"]

        cmd_mock = MagicMock(return_value=cmd_ret)
        with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
            os.environ, self.orig_env
        ):
            result = syslog_ng.config_test()
            self.assertEqual(result, cmd_ret)
            cmd_mock.assert_called_once_with(cmd_args, env=None, python_shell=False)

        cmd_mock = MagicMock(return_value=cmd_ret)
        with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
            os.environ, self.orig_env
        ):
            result = syslog_ng.config_test(syslog_ng_sbin_dir=self.bin_dir)
            self.assertEqual(result, cmd_ret)
            cmd_mock.assert_called_once_with(
                cmd_args, env=self.mocked_env, python_shell=False
            )

    @skipIf(salt.utils.platform.is_windows(), "Module not available on Windows")
    def test_config_test_cfgfile(self):
        cfgfile = "/path/to/syslog-ng.conf"
        cmd_ret = {"retcode": 1, "stderr": "Syntax error...", "stdout": ""}
        cmd_args = ["syslog-ng", "--syntax-only", "--cfgfile={}".format(cfgfile)]

        cmd_mock = MagicMock(return_value=cmd_ret)
        with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
            os.environ, self.orig_env
        ):
            self.assertEqual(syslog_ng.config_test(cfgfile=cfgfile), cmd_ret)
            cmd_mock.assert_called_once_with(cmd_args, env=None, python_shell=False)

        cmd_mock = MagicMock(return_value=cmd_ret)
        with patch.dict(syslog_ng.__salt__, {"cmd.run_all": cmd_mock}), patch.dict(
            os.environ, self.orig_env
        ):
            self.assertEqual(
                syslog_ng.config_test(syslog_ng_sbin_dir=self.bin_dir, cfgfile=cfgfile),
                cmd_ret,
            )
            cmd_mock.assert_called_once_with(
                cmd_args, env=self.mocked_env, python_shell=False
            )
